/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	GESIPC:					*/
/*	Take care on message exchange between	*/
/*	process.				*/
/*						*/
/************************************************/
#include	<sys/time.h>
#include	<stdio.h>
#include	<signal.h>
#include	<string.h>
#include	<time.h>
#include	<errno.h>
#include	<syslog.h>
#
#include	"utlprc.h"
#include	"lowtyp.h"
#include	"subprc.h"
#include	"subclm.h"
#include	"unijrl.h"
#include	"gesipc.h"

#define	MILNAN	1000000	/*millisec in nanosec	*/

#define	MINF	   1024	/*max uniq msg size	*/

/*INFTYP flags values				*/
#define	PWAIT	 0X0001	/*ipc waiting for data	*/
#define	PWAKE	 0X0002	/*Wake up subscribed	*/
#define	PBUSY	 0X0004	/*ipc waiting for data	*/
#define	QUEUE	 0X0008	/*standard msg queue	*/
#define	POOL	 0X0010	/*msg queue is a pool	*/

#define	MIVALCPT    100	/*minimal MSGNUM value	*/
#define	MXVALCPT 999999	/*maximun MSGNUM value	*/

typedef	struct	{
	pid_t pidref;	/*message origine	*/
	u_long ctxref;	/*message context number*/
	}MSGUID;

typedef	struct	{
	u_long stamp;	/*time stamp		*/
	int reply;	/*message is a reply	*/
	pid_t from;	/*message origine	*/
	pid_t pidref;	/*original sender ref	*/
	u_long ctxref;	/*message context ref	*/
	u_long msgnum;	/*current msgnum	*/
	char info[MINF];/*message contents	*/
	}MSGINF;

typedef struct sigaction SIGACT;
typedef	struct	{	/*msg to process info	*/
	int poolnum;	/*pool number		*/
	short flags;	/*INF flasg status	*/
	pid_t process;	/*process pid		*/
	pid_t pool;	/*channel pool number	*/
	u_int inqueue;	/*process in queue	*/
	SIGACT alarm;	/*previous alarm trappe	*/
	}INFTYP;

typedef	struct	{	/*Message control info	*/
	u_long msgnum;	/*current msg number	*/
	u_int maxmsg;	/*max message in area	*/
	u_int avail;	/*avail # of msg slot	*/
	u_int mxproc;	/*maximun subscribe	*/
	u_int known;	/*current # of process	*/
	INFTYP *inf;	/*accepted recipient	*/
	MSGINF *msgs;	/*message area		*/
	pid_t sloms[1];	/*msg pool num in slot	*/
	}BOXTYP;

/*

*/
/************************************************/
/*						*/
/*	Procedure to catch signal and do what is*/
/*	needed.					*/
/*						*/
/************************************************/
static void trpipcsignal(int sig)

{
switch (sig) {
  case SIGALRM  :
    break;
  default       :
    (void) syslog(LOG_INFO,"gesipc.c:trpipcsignal Unexpected "
			   "signal <%s> received\n",sys_siglist[sig]);
    break;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to increase the box msgnum	*/
/*	counter.,				*/
/*						*/
/************************************************/
static void increasecpt(BOXTYP *box)

{
box->msgnum++;
box->msgnum%=MXVALCPT;
if (box->msgnum==(u_long)0)
  box->msgnum=MIVALCPT;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to compose a MSGREF from	*/
/*	a MSGUID information, allocated memory	*/
/*	is dynamic and MUST be released later	*/
/*	on.					*/
/*						*/
/************************************************/
static char *mkmsgref(MSGUID *msguid)

{
char *msgref;

(void) asprintf(&msgref,"%d:%lu",msguid->pidref,msguid->ctxref);
return msgref;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract a MSGUID from	*/
/*	a  MSGREF information			*/
/*						*/
/************************************************/
static void getmsguid(char *msgref,MSGUID *msguid)

{
#define	OPGT	"gesipc.c:getmsguid,"

msguid->pidref=(pid_t)0;
msguid->ctxref=(u_long)0;
if (sscanf(msgref,"%d:%lu",&msguid->pidref,&msguid->ctxref)!=2)
  (void) clm_alert(0,"%s Unable to convert MSGREF <%s>",OPGT,msgref);
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to manage message slot within	*/
/*	the box.				*/
/*	purpose is to set a set of slot to a	*/
/*	specific pid.				*/
/*						*/
/************************************************/
static void markslots(BOXTYP *box,pid_t pool,u_int nbrslot,u_int *slotlst)

{
register u_int i;
register u_int *ptr;
register pid_t *psl;

for (i=0,psl=box->sloms,ptr=slotlst;i<nbrslot;i++,ptr++) {
  *(psl+*ptr)=pool;
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to manage message slot within	*/
/*	the box.				*/
/*	purpose is to mark slot as used by a	*/
/*	messages.				*/
/*						*/
/************************************************/
static int takeslots(BOXTYP *box,INFTYP *inf,TIMESPEC *timeout,
		     u_int nbrslot,u_int *slotlst)

{
int status;

status=false;
if (ipc_getshraccess((void *)box,timeout)==true) {
  (void) markslots(box,inf->pool,nbrslot,slotlst);
  inf->inqueue+=nbrslot;
  (void) ipc_dropshraccess((void *)box);
  status=true;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to sens a SIGALRM signal to	*/
/*	waiting process among the message pool,	*/
/*	If the process is not in a messages pool*/
/*	signal si sent to the process itself.	*/
/*						*/
/************************************************/
static void dowakeup(BOXTYP *box,INFTYP *inf)

{
if ((inf->flags&(POOL|QUEUE))==(short)POOL) {
  register unsigned int i;
  register short mask;
  register INFTYP *ptr;

  mask=QUEUE|PWAIT|PWAKE;
  for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
    if (ptr->pool==(pid_t)0)
      continue;
    if (ptr->pool==inf->pool) {
      if ((ptr->flags&mask)>QUEUE) {
        (void) kill(ptr->process,SIGALRM);
	/*waking UP ALL pooling process	*/
	/*No break;*/
	}
      } 
    }
  }
else {
  (void) kill(inf->process,SIGALRM);
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to manage message slot within	*/
/*	the box.				*/
/*	purpose is to release slot used by 	*/
/*	extracted message.			*/
/*						*/
/************************************************/
static int dropslots(BOXTYP *box,INFTYP *inf,u_int nbrslot,u_int *slotlst)

{
int status;
TIMESPEC timeout;

status=false;
timeout.tv_sec=3;	/*3 sec to drop slot!		*/
timeout.tv_nsec=0;
if (ipc_getshraccess((void *)box,&timeout)==true) {
  (void) markslots(box,(pid_t)0,nbrslot,slotlst);
  box->avail+=nbrslot;
  if (inf!=(INFTYP *)0) {
    inf->inqueue-=nbrslot;
    }
  (void) ipc_dropshraccess((void *)box);
  status=true;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to deduct a pool queue from	*/
/*	the current pid.			*/
/*						*/
/************************************************/
static INFTYP *mypool(BOXTYP *box,pid_t process)

{
INFTYP *pptr;
register INFTYP *ptr;
register unsigned int i;

pptr=(INFTYP *)0;
for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
  if (ptr->pool==(u_long)0)
    continue;
  if (ptr->process==process) {
    pptr=ptr;
    break;	/*the process pool	*/
    }
  }
return pptr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to find a pool queue among	*/
/*	already open queue.			*/
/*						*/
/************************************************/
static INFTYP *findpool(BOXTYP *box,pid_t pool)

{
INFTYP *pptr;
register INFTYP *ptr;
register unsigned int i;

pptr=(INFTYP *)0;
for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
  if (ptr->pool==(pid_t)0)
    continue;
  if (ptr->pool==pool) {
    pptr=ptr;
    if ((ptr->flags&(short)(POOL|QUEUE))==(short)POOL) 
      break;	/*the MAIN pool		*/
    }
  }
return pptr;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to check if a process		*/
/*	subscribed to a message pool		*/
/*						*/
/************************************************/
static INFTYP *issubscribed(BOXTYP *box,pid_t process)

{
INFTYP *ison;
register INFTYP *ptr;
register unsigned int i;


ison=(INFTYP *)0;
for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
  if (ptr==(INFTYP *)0)
    break;
  if (ptr->process==process) {
    ison=ptr; 
    break;
    }
  }
return ison;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to reserve a specific slot	*/
/*	amount among the shared memory area.	*/
/*	If not enouhg space is available, return*/
/*	an empty list.				*/
/*						*/
/************************************************/
static u_int *getslot(BOXTYP *box,u_long *msgnum,u_int nbrslot)

{
u_int *list;
TIMESPEC timeout;

list=(u_int *)calloc(nbrslot,sizeof(pid_t));
timeout.tv_nsec=0;
timeout.tv_sec=1;	/*wait 1 sec to get a slot	*/
if (ipc_getshraccess((void *)box,&timeout)==true) {
  register u_int found;

  found=0; 
  if (box->avail>=nbrslot) {
    register pid_t *slt;
    register u_int i;
    register u_int s;
   
    for (i=0,s=0,slt=box->sloms;(found<nbrslot)&&(i<box->maxmsg);i++,slt++) {
      if (*slt==(u_long)0) {
        *slt=(u_long)1;
        list[s]=i;
        s++;
        found++;
        }
      }
    box->avail-=found;
    box->msgnum++;
    *msgnum=box->msgnum;
    }
  (void) ipc_dropshraccess((void *)box);
  if (found==0) {
    (void) free(list);
    list=(u_int *)0;
    }
  }
return list;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to count reply waiting for	*/
/*	current process in the IPC message area	*/
/*						*/
/************************************************/
static u_int countreply(BOXTYP *box,u_int ready,u_int **msgrdy)

{
u_int found;
u_int *msglst;

found=(u_int)0;
msglst=(u_int *)0;
if (ready>0) {
  register pid_t curpid;
  register unsigned int current;
  register unsigned int i;
  register u_int *slots;

  curpid=getpid();
  current=0;
  msglst=(u_int *)calloc(ready,sizeof(u_int));
  for (i=0,slots=*msgrdy;i<ready;i++,slots++) {
    register MSGINF *ptr;

    ptr=box->msgs+*slots;
    if (ptr->reply==false)
      continue;
    if (ptr->msgnum==current)
      continue; 
    if (ptr->pidref!=curpid)
      continue;
    current=ptr->msgnum; 
    msglst[found]=*slots;
    found++; 
    }
  if (found==0) {
    (void) free(msglst);
    msglst=(u_int *)0;
    }
  }
(void) free(*msgrdy);
*msgrdy=msglst;
return found;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract msgref for each	*/
/*	reply found.				*/
/*						*/
/************************************************/
static char **listreply(BOXTYP *box,register u_int ready,register u_int *msglst)

{
char **lst;

lst=(char **)0;
if (ready>0) {
  register u_int i;

  lst=(char **)calloc(ready+1,sizeof(char *));
  for (i=0;i<ready;i++,msglst++) {
    register MSGINF *ptr;

    ptr=box->msgs+*msglst;
    (void) asprintf(&(lst[i]),"%d:%lu",ptr->pidref,ptr->ctxref);
    }
  }
return lst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return a message number 	*/
/*	addressed to the current process	*/
/*						*/
/************************************************/
static u_int spotonemsg(BOXTYP *box,MSGUID *msguid,u_int ready,u_int **msgrdy)

{
register u_int found;
register u_long msgnum;

found=(u_int)0;
msgnum=(u_long)0;
if (ready>0) {
  register int reply;
  register u_int *slots;
  register u_int *msglst;
  register u_int i;

  reply=false;
  if (msguid->ctxref!=0)
    reply=true;
  slots=*msgrdy;
  msglst=(u_int *)calloc(ready,sizeof(u_int));
  for (i=0;i<ready;i++,slots++) {
    register MSGINF *ptr;
    u_int another;

    another=false;
    ptr=box->msgs+*slots;
    if (ptr->reply!=reply)
      continue;		/*we are looking for reply or post	*/
    if (msgnum==(u_long)0) {
      if (msguid->ctxref==0) {
        msgnum=ptr->msgnum;
        another=true;
        }
      else {
        if (msguid->ctxref==ptr->ctxref) {
          msgnum=ptr->msgnum;
          another=true;
	  }
        }
      }
    else {
      if (msguid->ctxref==0) {
        if (ptr->msgnum==msgnum)
          another=true;
        else {
          if (ptr->msgnum<msgnum) {
	    another=true;
            msgnum=ptr->msgnum;
	    found=0;
	    }
	  }
        }
      else {
        if (msguid->ctxref==ptr->ctxref)
	  another=true;
        }
      }
    if (another==true) {
      msglst[found]=*slots;
      found++;
      }
    }
  if (found==0) {
    (void) free(msglst);
    msglst=(u_int *)0;
    }
  (void) free(*msgrdy);
  *msgrdy=msglst;
  }
return found;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract the list of slot	*/
/*	with a message ready to be grabed.	*/
/*						*/
/************************************************/
static u_int getready(BOXTYP *box,TIMESPEC *timeout,INFTYP *inf,
		      MSGUID *msguid,u_int **msglst)

{
#define	OPGR	"gesipc.c:getready,"

u_int ready;

ready=(u_int)0;
if (ipc_getshraccess((void *)box,timeout)==true) {
  register u_int max;

  max=inf->inqueue;
  if (max>0) {
    register pid_t *slt;
    register u_int *ptr;
    register pid_t pool;
    register u_int i;

    *msglst=(u_int *)calloc(max,sizeof(u_int));
    ptr=*msglst;    
    pool=inf->pool;
    for (i=0,slt=box->sloms;(ready<max)&&(i<box->maxmsg);i++,slt++) {
      if (*slt==pool) {
        *ptr=i;
	ptr++;	
        ready++;
        }
      }
    if ((ready=spotonemsg(box,msguid,ready,msglst))>0)
      (void) markslots(box,(u_long)1,ready,*msglst);
    }
  (void) ipc_dropshraccess((void *)box);
  if (ready>0) {
    register MSGINF *msg;

    msg=box->msgs+(*msglst[0]);
    msguid->pidref=msg->pidref;
    msguid->ctxref=msg->ctxref;
    }
  }
return ready;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to add a message to the 	*/
/*	message queue.				*/
/*	If the message can't be stored, errno is*/
/*	set to ENOMEM and a 0 status is 	*/
/*	returned.				*/
/*	if everything is right, procedure 	*/
/*	return a non zero value			*/
/*						*/
/************************************************/
static u_long sendmsg(BOXTYP *box,TIMESPEC *timeout,
		      MSGUID *msguid,pid_t pool,char *strmsg)

{
#define	OPIC	"gesipc.c:sendmsg,"

u_int *slots;
pid_t curpid;
u_int taille;
INFTYP *inf;
u_int needed;
u_long msgnum;
u_long start;

slots=(u_int *)0;
curpid=getpid();
taille=strlen(strmsg)+1;
needed=(taille/MINF)+1;
msgnum=(u_long)0;
if (needed>box->maxmsg) {
  (void) clm_alert(0,"%s Unable to post message <%s> to pool '%lu' "
		     "(error='message too big')",
		     OPIC,strmsg,pool);
  errno=EOVERFLOW;
  return msgnum;
  }
if ((inf=findpool(box,pool))==(INFTYP *)0) {
  (void) clm_alert(0,"%s Unable to post message <%s> to pool '%lu' "
		     "(error='process not a subscriber')",
		    OPIC,strmsg,pool);
  errno=ESRCH;
  return msgnum;
  }
start=clm_getmillisec();
while ((slots=getslot(box,&msgnum,needed))==(u_int *)0) {
  if (((clm_getmillisec()-start)/1000)>=(u_long)timeout->tv_sec) {
    (void) clm_alert(0,"%s Unable to post message <%s> to pool '%lu' "
		       "(error='No slot available')",
		       OPIC,strmsg,pool);
    errno=EBUSY;	/*no slots available	*/
    return msgnum;
    }
  /*lets relax, system really busy		*/
  (void) usleep(10000);	
  }
if (slots!=(u_int *)0) {
  register unsigned int i;
  MSGINF infloc;

  infloc.stamp=start;
  infloc.reply=false;
  infloc.from=curpid;
  infloc.msgnum=msgnum;
  infloc.ctxref=msgnum;
  infloc.pidref=curpid;
  if (msguid!=(MSGUID *)0) {
    infloc.reply=true;
    infloc.ctxref=msguid->ctxref;
    infloc.pidref=msguid->pidref;
    }
  for (i=0;i<needed;i++) {
    register MSGINF *msg;
    u_int max;

    max=taille;
    if (max>MINF)
      max=MINF;
    msg=box->msgs+slots[i];
    (void) memcpy(msg,&infloc,sizeof(MSGINF));
    (void) memcpy(msg->info,strmsg+(MINF*i),max);
    taille-=max;
    }
  if (takeslots(box,inf,timeout,needed,slots)==false) {
    (void) clm_alert(0,"%s Unable to post message=<%s> (error=timeout)",
			  OPIC,strmsg);
    }
  else {
    (void) dowakeup(box,inf);
    (void) clm_alert(10,"By '%05d' STORE IPC [%05d:%lu:%lu] msg <%s>",
                         getpid(),infloc.pidref,infloc.ctxref,msgnum,strmsg);
    }
  (void) free(slots);
  }
else {
  (void) clm_alert(0,"%s Unable to send message=<%s> needed='%d', "
		     "current available='%d'",OPIC,strmsg,needed,box->avail);
  }
return msgnum;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to extract data contents and	*/
/*	remove message with a specific message	*/
/*	number from the shared area.		*/
/*	return for how many millisec the message*/
/*	stayed within the queue			*/
/*						*/
/************************************************/
static char *extract(BOXTYP *box,INFTYP *inf,u_int found,
		     u_int *lst,u_long *msec)

{
char *got;
register MSGINF *ptr;
register u_int i;

got=(char *)calloc(found,MINF);
ptr=box->msgs;
*msec=clm_getmillisec()-ptr->stamp;
for (i=0;i<found;i++) {
  register MSGINF *data;

  data=ptr+lst[i];
  (void) memcpy(got+(MINF*i),data->info,MINF);
  }
if (dropslots(box,inf,found,lst)==false) {
  (void) free(got);
  got=(char *)0;
  }
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to locate a message a return	*/
/*	the message contents en removing it	*/
/*	from the message queue.			*/
/*						*/
/************************************************/
static char *locate(BOXTYP *box,TIMESPEC *timeout,INFTYP *inf,
		    MSGUID *msguid,u_long *msec)

{
char *got;
u_int *msglst;
u_int ready;

got=(char *)0;
msglst=(u_int *)0;
if ((ready=getready(box,timeout,inf,msguid,&msglst))>0) {
  got=extract(box,inf,ready,msglst,msec);
  if ((debug>=10)&&(got!=(char *)0)) {
    u_long msgnum;
   
    msgnum=(box->msgs+msglst[0])->msgnum;
    (void) clm_alert(10,"By '%05d' GOT IPC [%05d:%lu:%lu] msg <%s>",
                         getpid(),msguid->pidref,msguid->ctxref,
                         msgnum,got);
    }
  }
if (msglst!=(u_int *)0)
  (void) free(msglst);
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove a message from the 	*/
/*	message queue.				*/
/*	If a message is not available within	*/
/*	time-out elapse time, a null pointer is	*/
/*	returned.				*/
/*						*/
/************************************************/
static char *getmsg(BOXTYP *box,TIMESPEC *timeout,MSGUID *msguid,u_long *msec)

{
#define	OPGE	"gesipc.c:getmsg,"

char *got;
pid_t curpid;
INFTYP *chan;
INFTYP *inf;

got=(char *)0;
curpid=getpid();
if ((chan=mypool(box,curpid))==(INFTYP *)0)
  (void) clm_core_dump("%s need to subscribe before getting first message",
			OPGE);
inf=findpool(box,chan->pool);	/*find pool if exist	*/
while (got==(char *)0) {
  int wake;

  if (inf->inqueue>0) {
    if ((got=locate(box,timeout,inf,msguid,msec))!=(char *)0) {
      break;
      }
    }
  chan->flags|=PWAIT;
  wake=nanosleep(timeout,timeout);
  chan->flags&=~PWAIT;
  if (wake==0)
    break;
  }
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to add a process number to the*/
/*	list of known process number.		*/
/*						*/
/************************************************/
static int dosubscribe(BOXTYP *box,u_long pool)

{
#define	OPSU	"gesipc.c:dosubscribe,"

int done;

done=false;
if (box->known<box->mxproc) {
  pid_t curpid;

  curpid=getpid();
  if (issubscribed(box,curpid)==(INFTYP *)0) {
    TIMESPEC timeout;

    timeout.tv_nsec=100*MILNAN;	/*100 msec	*/
    timeout.tv_sec=0;
    if (ipc_getshraccess((void *)box,&timeout)==true) {
      register INFTYP *inf;

      if ((inf=issubscribed(box,(pid_t)0))!=(INFTYP *)0) {
        struct sigaction newsa;

        newsa.sa_flags=0;
        newsa.sa_handler=trpipcsignal;
        (void) sigemptyset(&newsa.sa_mask);
        (void) sigaction(SIGALRM,&newsa,&(inf->alarm));
        box->known++;
        inf->inqueue=(u_int)0;
        inf->flags=(short)QUEUE;
        inf->pool=pool;
	if (pool==(u_long)0) {
	  inf->pool=box->msgnum;	/*setting a pool number	*/
	  inf->flags|=POOL;
          (void) increasecpt(box);
	  }
        inf->process=curpid;
        done=true;
	}
      (void) ipc_dropshraccess((void *)box);
      }
    }
  else
    (void) clm_alert(0,"%s pid '%05d' alreay in known list!",OPSU,curpid);
  }
else {
  (void) clm_alert(0,"%s pid='%d' too many process (%d) already subscribed!",
		     OPSU,getpid(),box->known);
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove a process number	*/
/*	from the list of known process number.	*/
/*						*/
/************************************************/
static int dounsubscribe(BOXTYP *box,pid_t curpid)

{
#define	OPUS	"gesipc.c:dounsubscribe,"

int done;
INFTYP *inf;

done=false;
if ((inf=issubscribed(box,curpid))!=(INFTYP *)0) {
  register u_long pool;
  TIMESPEC timeout;

  pool=inf->pool;
  timeout.tv_nsec=0;	/*1 secondes		*/
  timeout.tv_sec=1;
  if (ipc_getshraccess((void *)box,&timeout)==true) {
    register unsigned int i;
    register pid_t *ptr;
    register MSGINF *msg;

    for (i=0,ptr=box->sloms,msg=box->msgs;i<box->maxmsg;i++,ptr++,msg++) {
      if (*ptr==(pid_t)pool) {
	*ptr=(pid_t)0;
	}
      }
    inf->pool=(pid_t)0;
    inf->process=(pid_t)0;
    (void) sigaction(SIGALRM,&(inf->alarm),(SIGACT *)0);
    box->known--;
    (void) ipc_dropshraccess((void *)box);
    done=true;
    }
  }
else {
  (void) clm_alert(0,"%s pid '%05d' not in known list!",OPUS,curpid);
  }
return done;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to assign a shared messages	*/
/*	area used to exchange message between	*/
/*	2 clement process.			*/
/*	where mxksize is the number of KiloByte	*/
/*	dedicated to shared memory and		*/
/*	mxproc is the maximun number of message	*/
/*	in queue.				*/
/*						*/
/************************************************/
BPTR *ipc_getbox(u_int mxksize,u_int mxproc)

{
#define	OPGB	"gesipc.c:ipc_getbox,"

BOXTYP *box;
u_int taille;
u_int boxsize;
int phase;
int proceed;

box=(BOXTYP *)0;
boxsize=(mxksize*1024)/sizeof(MSGINF);
taille=sizeof(BOXTYP)+(boxsize*sizeof(pid_t));
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*Trouble collector	*/
      if (box->inf!=(INFTYP *)0)
        box->inf=(INFTYP *)ipc_dropshrmem((void *)box->inf);
      if (box->msgs!=(MSGINF *)0)
        box->msgs=(MSGINF *)ipc_dropshrmem((void *)box->msgs);
      box=(BOXTYP *)ipc_dropshrmem((void *)box);
      break;
    case 1	:	/*getting the BOX	*/
      if ((box=(BOXTYP *)ipc_getshrmem(1,taille))==(BOXTYP *)0) {
 	proceed=false;	/*trouble Trouble	*/
	}
      break;
    case 2	:	/*preparing Box		*/
      box->maxmsg=boxsize;
      box->avail=boxsize;
      box->msgnum=MIVALCPT;
      box->mxproc=mxproc;
      break;
    case 3	:	/*adding shared msg area*/
      box->msgs=(MSGINF *)ipc_getshrmem(boxsize,sizeof(MSGINF));
      if (box->msgs==(MSGINF *)0) {
        (void) clm_alert(0,"%s Unable to get message space",OPGB);
	phase=-1;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*adding shared info	*/
      box->inf=(INFTYP *)ipc_getshrmem(mxproc,sizeof(INFTYP));
      if (box->inf==(INFTYP *)0) {
        (void) clm_alert(0,"%s Unable to get message info",OPGB);
	phase=-1;	/*trouble trouble	*/
	}
      break;
    case 5	:	/*marking inf slot	*/
      if (box->inf!=(INFTYP *)0) {	/*always*/
	unsigned int i;

	for (i=0;i<mxproc;i++) {
	  (box->inf+i)->poolnum=i;
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;	/*everything fine	*/
      break;
    }
  phase++;
  }
return (BPTR *)box;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to free a shared messages	*/
/*	area.					*/
/*						*/
/************************************************/
BPTR *ipc_dropbox(BPTR *bptr)

{
BOXTYP *box;

box=(BOXTYP *)bptr;
if (box!=(BOXTYP *)0) {
  box->inf=(INFTYP *)ipc_dropshrmem((void *)box->inf);
  box->msgs=(MSGINF *)ipc_dropshrmem((void *)box->msgs);
  box=(BOXTYP *)ipc_dropshrmem((void *)box);
  }
return (BPTR *)box;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to allow message sender	*/
/*	process to wake up curent process	*/
/*						*/
/************************************************/
void ipc_wakemeup(BPTR *bptr,int onoff)

{
register BOXTYP *box;
register INFTYP *inf;

box=(BOXTYP *)bptr;
if ((inf=issubscribed(box,getpid()))!=(INFTYP *)0) {
  switch (onoff) {
    case true		:
      inf->flags|=PWAKE;
      break;
    default		:
      inf->flags&=~PWAKE;
      break;
    }
  }
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to subscribe/unsubscribe	*/
/*	a process to message exchange pool	*/
/*						*/
/************************************************/
int ipc_subscribe(BPTR *bptr,u_long pool,int onoff)

{
int status;
BOXTYP *box;

status=false;
box=(BOXTYP *)bptr;
if (box!=(BOXTYP *)0) {
  if (onoff==true)
    status=dosubscribe(box,pool);
  else 
    status=dounsubscribe(box,getpid());
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the channel pool	*/
/*	number linked to a specific process	*/
/*						*/
/************************************************/
u_long ipc_getpool(BPTR *bptr,pid_t process)

{
u_long pool;
BOXTYP *box;
INFTYP *inf;

pool=(u_long)0;
if (bptr==(BPTR *)0)
  return pool;
box=(BOXTYP *)bptr;
if ((inf=issubscribed(box,process))!=(pid_t)0) {
  pool=inf->pool;
  }
return pool;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to define a new pool id	*/
/*	area.					*/
/*						*/
/************************************************/
u_long ipc_openpool(BPTR *bptr)

{
#define	OPPL	"gesipc.c:openpool"

u_long pool;
BOXTYP *box;
TIMESPEC timeout;

pool=(u_long)0;
box=(BOXTYP *)bptr;
timeout.tv_nsec=100*MILNAN;	/*100 msec	*/
timeout.tv_sec=0;
if (ipc_getshraccess((void *)box,&timeout)==true) {
  INFTYP *inf;

  if ((inf=issubscribed(box,(pid_t)0))!=(INFTYP *)0) {
    pool=box->msgnum;
    inf->flags=POOL;
    inf->pool=pool;
    inf->process=(pid_t)1;
    inf->inqueue=(u_int)0;
    box->known++;
    (void) increasecpt(box);
    }
  (void) ipc_dropshraccess((void *)box);
  if (pool==(u_long)0)
    (void) clm_alert(0,"%s, Unable to find a POOL space",OPPL);
  }
return pool;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to close a message pool and	*/
/*	unlink all reference to other process	*/
/*						*/
/************************************************/
u_long ipc_closepool(BPTR *bptr,u_long pool)

{
BOXTYP *box;
TIMESPEC timeout;

box=(BOXTYP *)bptr;
timeout.tv_nsec=100*MILNAN;	/*100 msec	*/
timeout.tv_sec=0;
if (pool!=(u_long)0) {
  if (ipc_getshraccess((void *)box,&timeout)==true) {
    int done;
    register INFTYP *ptr;
    register unsigned int i;

    done=false;
    for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
      if (ptr->pool==(pid_t)pool) {
        switch (ptr->flags&(POOL|QUEUE))	{
	  case POOL		:
	    ptr->process=(pid_t)0;
	    ptr->flags=0;
	    ptr->inqueue=0;
	    done=true;
	    break;	
	  case QUEUE		:
	    break;	
	  case POOL|QUEUE	:
	    ptr->flags=QUEUE;
	    ptr->pool=(u_long)0;
	    break;	
	  }
        }
      }
    (void) ipc_dropshraccess((void *)box);
    if (done==true)
      pool=(u_long)0;
    }
  }
return pool;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to list all reply message	*/
/*	pending in the message area.		*/
/*						*/
/************************************************/
char **ipc_listreply(BPTR *bptr,TIMESPEC *timeout)

{
#define	OPLI	"gesipc.c:ipc_listreply,"

char **lst;
BOXTYP *box;
INFTYP *chan;
INFTYP *inf;

lst=(char **)0;
box=(BOXTYP *)bptr;
if ((chan=mypool(box,getpid()))==(INFTYP *)0)
  (void) clm_core_dump("%s need to subscribe before getting a message list",OPLI);
inf=findpool(box,chan->pool);   /*find pool if exist    */
if (ipc_getshraccess((void *)box,timeout)==true) {
  register u_int ready;
  register u_int max;
  u_int *msglst;

  ready=(u_int)0;
  max=inf->inqueue;
  msglst=(u_int *)calloc(max,sizeof(u_int));
  if (max>0) {
    register pid_t *slt;
    register u_int *ptr;
    register pid_t pool;
    register u_int i;

    ptr=msglst;
    pool=inf->pool;
    for (i=0,slt=box->sloms;(ready<max)&&(i<box->maxmsg);i++,slt++) {
      if (*slt==pool) {
        *ptr=i;
        ptr++;
        ready++;
        }
      }
    ready=countreply(box,ready,&msglst);
    }
  (void) ipc_dropshraccess((void *)box);
  lst=listreply(box,ready,msglst);
  (void) free(msglst);
  }
return lst;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to post a message to a 	*/
/*	specific process.			*/
/*	If the message can't be sent, function	*/
/*	return a NULL pointer as message	*/
/*	reference.				*/
/*						*/
/************************************************/
char *ipc_postmsg(BPTR *bptr,TIMESPEC *timeout,u_long pool,char *msg)

{
#define	OPPO	"gesipc.c:ipc_postmsg,"

char *msgref;
BOXTYP *box;
u_long msgnum;

msgref=(char *)0;
box=(BOXTYP *)bptr;
msgnum=0;
if ((msgnum=sendmsg(box,timeout,(MSGUID *)0,pool,msg))!=0) {
  MSGUID msguid;

  msguid.pidref=getpid();
  msguid.ctxref=msgnum;
  msgref=mkmsgref(&msguid);
  }
return msgref;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get the first message 	*/
/*	for the current procees available	*/
/*	in the queue.				*/
/*	If a message is avaliable return the	*/
/*	message, the message reference and the	*/
/*	stamping date.				*/
/*						*/
/************************************************/
char *ipc_getpost(BPTR *bptr,TIMESPEC *timeout,
		  char **msgref,u_long *millisec)

{
char *msg;
BOXTYP *box;
MSGUID msguid;

msg=(char *)0;
box=(BOXTYP *)bptr;
msguid.ctxref=(u_long)0;
if ((msg=getmsg(box,timeout,&msguid,millisec))!=(char *)0) {
  *msgref=mkmsgref(&msguid);
  } 
return msg;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to post a reply to a 		*/
/*	previously get message.			*/
/*	Return zero if trouble otherwise the	*/
/*	message number.				*/
/*						*/
/************************************************/
u_long ipc_postreply(BPTR *bptr,TIMESPEC *timeout,char *msgref,char *msg)

{
u_long msgnum;
u_long pool;
BOXTYP *box;
MSGUID msguid;

msgnum=(u_long)0;
box=(BOXTYP *)bptr;
(void) getmsguid(msgref,&msguid);
if ((pool=ipc_getpool(bptr,msguid.pidref))!=(u_long)0) {
  msgnum=sendmsg(box,timeout,&msguid,pool,msg);
  }
return msgnum;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to get a specific reply to	*/
/*	a previously sent message.		*/
/*						*/
/************************************************/
char *ipc_getreply(BPTR *bptr,TIMESPEC *timeout,char *msgref)

{
BOXTYP *box;
MSGUID msguid;
u_long locmilli;
char *got;

box=(BOXTYP *)bptr;
(void) getmsguid(msgref,&msguid);
locmilli=(u_long)0;
got=getmsg(box,timeout,&msguid,&locmilli);
return got;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove lost Message within	*/
/*	The queue.				*/
/*	ellapse is the time in seconde between	*/
/*	2 scan.					*/
/*						*/
/************************************************/
int ipc_cleanbox(BPTR *bptr,u_int ellapse)

{
static u_long last=(u_long)0;

int cleandone;
u_long millitime;

cleandone=false;
ellapse*=1000;	/*ellapse is in second		*/
millitime=clm_getmillisec();
if ((last+ellapse)<millitime) {
  register BOXTYP *box;
  register INFTYP *ptr;
  register unsigned int i;

  box=(BOXTYP *)bptr;
  for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
    if ((ptr->pool==(u_long)0)||(ptr->process==1))
      continue;
    if (prc_checkprocess(ptr->process)==false) {
      (void) jrl_journalling(true,APJRL,"Process '%05d' lost, "
					"now removing message "
					"to process from the IPC box",
				         ptr->process);
      (void) dounsubscribe(bptr,ptr->process);
      }
    }
  for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
    if (box->sloms[i]<MIVALCPT)
      continue;
    if ((box->msgs[i].stamp+(2*ellapse))<millitime) {
      register INFTYP *inf;

      if ((inf=findpool(box,box->sloms[i]))!=(INFTYP *)0) {
	if (inf->inqueue>0) {
	  const char *msgfrm;
          TIMESPEC timeout;
	  MSGUID msguid;
	  u_long stamp;
	  char *got;

	  msgfrm="Removing (from IPC) a too old message <Age: %ldms, Msgref: %s --> %.*s>%s";
  	  timeout.tv_sec=0;
  	  timeout.tv_nsec=200*MILNAN;
	  msguid.pidref=(pid_t)0;
	  msguid.ctxref=box->msgs[i].ctxref;
          box->msgs[i].reply=true;
	  if ((got=locate(box,&timeout,inf,&msguid,&stamp))!=(char *)0) {
	    char *msgref;

	    msgref=mkmsgref(&msguid);
            (void) jrl_journalling(true,APJRL,msgfrm,stamp,msgref,MINF,got,"");
	    (void) clm_alert(0,msgfrm,stamp,msgref,MINF,got," (bug?)");
	    (void) free(msgref);
	    (void) free(got);
	    }
	  }
	}
      }
    }
  cleandone=true;
  last=millitime;
  }
return cleandone;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the number of post	*/
/*	available to collect.			*/
/*						*/
/************************************************/
u_int ipc_postpending(BPTR *bptr)

{
u_int pending;
INFTYP *chan;
INFTYP *inf;

pending=0;
if ((chan=mypool((BOXTYP *)bptr,getpid()))!=(INFTYP *)0) {
  inf=findpool((BOXTYP *)bptr,chan->pool);	/*find pool if exist	*/
  if (inf!=(INFTYP *)0)
    pending=inf->inqueue;
  }
return pending;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return true if box is 	*/
/*	empty enough to accept a message post	*/
/*						*/
/************************************************/
int ipc_boxready(BPTR *bptr)

{
int ready;
BOXTYP *box;

ready=true;
box=(BOXTYP *)bptr;
if (box->avail<(box->maxmsg/4)) 
  ready=false;
return ready;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to return the number of 	*/
/*	messages still pending in the queue	*/
/*	FROM the current process.		*/
/*						*/
/************************************************/
u_int ipc_mypending(BPTR *bptr)

{
register u_int pending;
register pid_t curpid;
register BOXTYP *box;
register unsigned int i;
register MSGINF *msg;
register pid_t *ptr;

pending=0;
curpid=getpid();
box=(BOXTYP *)bptr;
for (i=0,ptr=box->sloms,msg=box->msgs;i<box->maxmsg;i++,ptr++,msg++) {
  if (*ptr==(pid_t)0)
    continue;
  if (msg->from==curpid)
    pending++;
  }
return pending;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Procedure to remove a message posted	*/
/*	previously but we are NOT waiting for	*/
/*	an answer anymore.			*/
/*	if the message was not found, return 	*/
/*	false, true otherwise.			*/
/*						*/
/************************************************/
u_int ipc_rmmymsg(BPTR *bptr,TIMESPEC *timeout,char *msgref)

{
int status;
BOXTYP *box;
MSGUID msguid;
register INFTYP *ptr;
register unsigned int i;

status=false;
box=(BOXTYP *)bptr;
(void) getmsguid(msgref,&msguid);
for (i=0,ptr=box->inf;i<box->mxproc;i++,ptr++) {
  if (box->sloms[i]<MIVALCPT)
    continue;
  if (msguid.ctxref==box->msgs[i].ctxref) {
    INFTYP *inf;


    if ((inf=findpool(box,box->sloms[i]))!=(INFTYP *)0) {
      u_long stamp;
      char *got;

      box->msgs[i].reply=true;
      msguid.pidref=getpid();
      if ((got=locate(box,timeout,inf,&msguid,&stamp))!=(char *)0) {
	(void) free(got);
	status=true;
	break;
	}
      }
    }
  }
return status;
}
