/************************************************/
/*						*/
/*      Copyright:				*/
/*		 Jean-Marc Pigeon <jmp@safe.ca>	*/
/*	Distributed under the Gnu Public	*/
/*	License, see the License file in this	*/
/*	package.				*/
/*						*/
/*	LVLEML:					*/
/*	Management level routines to handle	*/
/*	SMTP Protocol exchange with remote site.*/
/*						*/
/************************************************/
#include	<sys/file.h>
#include	<sys/types.h>
#include	<arpa/inet.h>
#include	<fcntl.h>
#include	<netdb.h>
#include	<malloc.h>
#include	<string.h>
#include	<errno.h>
#include	<ctype.h>
#include	<time.h>
#
#include	"lowtyp.h"
#include	"utlvec.h"
#include	"subclm.h"
#include	"subprc.h"
#include	"bascnv.h"
#include	"unijrl.h"
#include	"unidns.h"
#include	"gesclm.h"
#include	"comeml.h"
#include	"lvleml.h"

#define MXTIME	120	/*wait for a 300 sec max*/
#define	CRLF	"\r\n"

typedef	enum	{	/*SMTP vocabulary code	*/
	c_help,		/*provide help		*/
	c_mail,		/*mail from		*/
	c_rcpt,		/*recipient		*/
	c_ehlo,		/*code extension	*/
	c_helo,		/*code presentation	*/
	c_data,		/*data mode		*/
	c_tls,		/*Transport Socket Layer*/
	c_auth,		/*user authentication	*/
	c_reset,	/*reset current stat	*/ 
	c_noop,		/*operation NOOP	*/
	c_quit,		/*abort connection	*/
	c_chkid,	/*user checking ID	*/
	c_ckid,		/*to get site dfltcode	*/
	c_trusting,	/*origin is trusted	*/
	c_orgn,		/*origin override	*/
	c_unknown	/*unknown code		*/
	}CODTYP;

typedef	struct	{	/*available command	*/
	CODTYP code;	/*test code		*/	
	SCODTYP scode;	/*related smtp code	*/	
	int proceed;	/*continue to proceed	*/
	char **help;	/*command explainations	*/
	}ACTTYP;

static	char *hlphelo[]={
		"HELO <hostname>",
		"\tIntroduce yourself.",
		(char *)0
		};

static	char *hlpehlo[]={
		"EHLO <hostname>",
		"\tIntroduce yourself and request extended SMTP mode.",
		"\t8BITMIME\tUse 8-bit data\t\t[RFC1652]",
		"\tSIZE\tMessage size declaration\t\t[RFC1870]",
		"\tSTARTTLS\tSecure SMTP\t\t[RFC2487]",
		"\tTRUSTING\tRemote site is seen as a trusted and can send banned remote list to the remote connection",
		(char *)0
		};

static	char *hlpmail[]={
		"MAIL FROM: <sender> [ <parameters> ]",
		"\tSpecifies the sender.  Parameters are ESMTP extensions.",
		(char *)0
		};

static	char *hlprcpt[]={
		"RCPT TO: <recipient> [ <parameters> ]",
		"\tSpecifies the recipient. Can be used any number of times.",
		"\tParameters are ESMTP extensions.",
		(char *)0
		};

static	char *hlpdata[]={
		"DATA",
		"\tFollowing text is collected as the message.",
		"\tEnd with a single dot.",
		(char *)0
		};

static	char *hlpreset[]={
		RSET,
		"\tResets the system.",
		(char *)0
		};

static	char *hlpnoop[]={
		STRNOOP,
		"\tDo nothing.",
		(char *)0
		};

static	char *hlpquit[]={
		SQUIT,
		"\tExit sendmail (SMTP).",
		(char *)0
		};

static	char *hlphelp[]={
		"HELP [ <topic> ]",
		"\tThe HELP command gives help info.",
		(char *)0
		};

static	char *hlptls[]={
		"STARTTLS",
		"\tStart TLS negotiation.",
		(char *)0
		};

static	char *hlpauth[]={
		"AUTH mechanism [initial-response]",
		"\tStart authentication.",
		(char *)0
		};

static	char *hlpchkid[]={
		"CHECKID <userid> password uniqid",
		"\tto check a user id and provide further "
		"access to clement WEB interface",
		(char *)0
		};

static	char *hlpckid[]={
		"CKID dfltcode",
		"\tsend the remote system authentication",
		(char *)0
		};

static	char *hlptrusted[]={
		"TRUSTING",
		"\tFollowing text is collected as formated trusting data.",
		"\tEnd with a single dot.",
		(char *)0
		};

static	char *hlporgn[]={
		"ORGN <ipnumber>",
		"\tOverride the remote detection (debugging only)",
		(char *)0
		};

static ACTTYP action[]={ /*MUST in CODTYP order	*/
		{c_help,s_unknown,true,hlphelp},
		{c_mail,s_mail,false,hlpmail},
		{c_rcpt,s_rcpt,false,hlprcpt},
		{c_ehlo,s_ehlo,false,hlpehlo},
		{c_helo,s_helo,false,hlphelo},
		{c_data,s_data,false,hlpdata},
		{c_tls,s_unknown,true,hlptls},
		{c_auth,s_auth,false,hlpauth},
		{c_reset,s_reset,false,hlpreset},
		{c_noop,s_unknown,true,hlpnoop},
		{c_quit,s_quit,false,hlpquit},
		{c_chkid,s_chkid,false,hlpchkid},
		{c_ckid,s_id,false,hlpckid},
		{c_trusting,s_trusting,true,hlptrusted},
		{c_orgn,s_unknown,true,hlporgn},
		{c_unknown,s_unknown,true,(char **)0}
		};

static VOCTYP vocsmtp[]={
		{c_helo,"HELO",(const void *)(action+c_helo)},
		{c_ehlo,"EHLO",(const void *)(action+c_ehlo)},
		{c_mail,MAILF,(const void *)(action+c_mail)},
		{c_rcpt,RCPT,(const void *)(action+c_rcpt)},
		{c_data,DATA,(const void *)(action+c_data)},
		{c_tls,"STARTTLS",(const void *)(action+c_tls)},
		{c_auth,SAUTH,(const void *)(action+c_auth)},
		{c_reset,RSET,(const void *)(action+c_reset)},
		{c_noop,STRNOOP,(const void *)(action+c_noop)},
		{c_quit,SQUIT,(const void *)(action+c_quit)},
		{c_help,"HELP",(const void *)(action+c_help)},
		{c_trusting,HLOTRS,(const void *)(action+c_trusting)},
		{c_orgn,"ORGN",(const void *)(action+c_orgn)},
		{c_unknown,(char *)0,(const void *)0}
		};

/*defining email part extension			*/
#define	HDPART	0
#define	DTPART	1
#define	TRPART	2
static	char *strpart[]={
	"-hder",	/*email header extension*/
	"-data",	/*email data extension	*/
	"-trust",	/*trust data extension	*/
	(char *)0
	};
/*

*/
/************************************************/
/*						*/
/*	Procedure to return a unique IMAP	*/
/*	cequence string build arround a number	*/
/*						*/
/************************************************/
static const char *getimapseq(u_int number)

{
static char seq[10];

(void) snprintf(seq,sizeof(seq),"C%04d",number);
return seq;
}
/*

*/
/************************************************/
/*						*/
/*	copying on file content to another file	*/
/*						*/
/************************************************/
static int copyfile(FILE *dst,FILE *src,char *srcname)

{
#define	OPCY	"lvleml.c:copyfile,"

int doneok;

doneok=false;
if ((src!=(FILE *)0)&&(dst!=(FILE *)0)) {
  int nbr;
  char buffer[2048];

  while ((nbr=fread(buffer,sizeof(char),sizeof(buffer),src))>0) 
    (void) fwrite(buffer,sizeof(char),nbr,dst);
  doneok=true;
  if ((feof(src)==0)&&(ferror(src)!=0)) {
    doneok=false;
    (void) clm_alert(0,"%s Unable to fully read <%s> (error=<%s>)",OPCY,
		        srcname,strerror(ferror(src)));
    }
  }
return doneok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report an input trouble	*/
/*	both to	remote server AND journal logs	*/
/*						*/
/************************************************/
static void baddatareport(CTTYP *contact,char *sequence,char *buffer,int status,char *cmd)

{
#define	ADDFRM	"\n\t%*s%s"

char *info;
char *curlog;

(void) asprintf(&curlog,"%s: %s",sequence,buffer);
info=eml_ascstat(status,cmd);
(void) eml_printfcrlf(contact,info);
(void) clm_fasprintf(&curlog,ADDFRM,2,"",info);
(void) free(info);
(void) jrl_journalling(false,EVJRL,"%s",curlog);
(void) free(curlog);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report SID information 	*/
/*	on journal				*/
/*						*/
/************************************************/
static void tracesid(CTTYP *contact,int reset)

{
char ext[100];

if (reset==false)
  (void) strcpy(ext,"-> Contact open");
else
  (void) snprintf(ext,sizeof(ext),"<- %s",contact->msession);
(void) jrl_journalling(false,EVJRL,"%-*s%s %s",6,SESSION,contact->session,ext);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report a syntax error 	*/
/*	both to remote contact and in logs.	*/
/*						*/
/************************************************/
static void synerr(CTTYP *contact,char *buffer)

{
(void) eml_report(contact,SYNERR,buffer);
(void) jrl_journalling(false,EVJRL,"Unexpected string sequence <%s>",buffer);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to skip after colon within	*/
/*	the SMTP protocol syntax		*/
/*						*/
/************************************************/
static int checkcolon(CTTYP *contact,char *buffer)

{
int itsok;
char *ptr;

itsok=true;

if ((ptr=strchr(buffer,':'))==(char *)0) {
  (void) synerr(contact,buffer);
  itsok=false;
  }
else {
  ptr++;
  (void) clm_skipblank(ptr);
  (void) memmove(buffer,ptr,strlen(ptr)+1);
  }
return itsok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to count the number of	*/
/*	recipient in queue and entitled to 	*/
/*	receive data E-mail.			*/
/*						*/
/************************************************/
static int countrcpts(RCPTYP **rcpts)

{
int count;

count=0;
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    switch (rcpts[i]->status) {
      case m_indata	:
      case m_acceptable	:
      case m_relayed	:
      case m_toforward	:
      case m_mustfwd	:
      case m_toresend	:
      case m_toresolv	:
	count++;
	break;
      default		:
        break;
       }
    }
  }
return count;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display connection status	*/
/*	if doing command 'ps ax' from console	*/
/*						*/
/************************************************/
static void emlsettitle(CTTYP *contact)

{
(void) prc_settitle("%s: Reading on %s:%s since %s, from '%s/[%s]'",
                    appname,contact->locip,contact->servport,contact->attime,
		    contact->peername,contact->peerip);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove any memory used by	*/
/*	an from user information structure	*/
/*						*/
/************************************************/
static FRMTYP *cleanfrm(FRMTYP *frm)

{
if (frm!=(FRMTYP *)0) {
  frm->stainfo=clm_freestr(frm->stainfo);
  frm->efrom=clm_freestr(frm->efrom);
  frm->mfrom=clm_freestr(frm->mfrom);
  frm->locip=clm_freestr(frm->locip);
  frm->servport=clm_freestr(frm->servport);
  frm->peerip=clm_freestr(frm->peerip);
  (void) free(frm);
  frm=(FRMTYP *)0;
  }
return frm;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove any memory used by 	*/
/*	a reciptiens information list		*/
/*						*/
/************************************************/
static RCPTYP **cleanrcpts(RCPTYP **rcpts)

{
if (rcpts!=(RCPTYP **)0) {
  register int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    rcpts[i]=eml_freercpt(rcpts[i]);
    }
  (void) free(rcpts);
  rcpts=(RCPTYP **)0;
  }
return rcpts;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to close an SMTP session	*/
/*						*/
/************************************************/
static void closesession(CTTYP *contact)

{
if (unlink(contact->mailname)<0) 
  (void) clm_alert(0,"%s Unable to unlink <%s> (error=<%s>)",
		     "lvleml.c:closesession",contact->mailname,
		     strerror(errno));
contact->clamvers=clm_freestr(contact->clamvers);
contact->mailname=clm_freestr(contact->mailname);
if (contact->dbgname!=(char *)0) {
  char *strloc;

  (void) asprintf(&strloc,"Closing session '%s'\n",contact->session);
  (void) tls_traceout(contact->tls,"#--",strloc,strlen(strloc));
  (void) free(strloc);
  }
contact->mtaille=(u_long)0;
contact->founddata=false;
contact->esmtp=false;
contact->info=clm_freestr(contact->info);
contact->session=clm_freestr(contact->session);
contact->virname=clm_freestr(contact->virname);
contact->rjctnote=clm_freestr(contact->rjctnote);
contact->evl=hdr_freeevl(contact->evl);
contact->headers=(char **)vec_freelstlst((void **)contact->headers,VECFREE);
contact->totbad+=contact->badrcpt;
contact->badrcpt=0;
contact->goodrcpt=0;
contact->foremail=clm_freestr(contact->foremail);
contact->rcpts=cleanrcpts(contact->rcpts);
contact->frm=cleanfrm(contact->frm);
contact->lastdata=(time_t)0;
contact->sequence++;
contact->sequence%=1000;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open an SMTP session	*/
/*						*/
/************************************************/
static void opensession(CTTYP *contact,int reset)

{
#define	OPOP	"lvleml.c:opensession,"


int queued;

if (contact->session==(char *)0) {
  char *strloc;

  contact->session=clm_uniquename(contact->sequence);
  (void) asprintf(&strloc,"Opening new session '%s'\n",contact->session);
  (void) tls_traceout(contact->tls,"#--",strloc,strlen(strloc));
  (void) free(strloc);
  }
(void) asprintf(&(contact->mailname),"%s/%s",clm_appdir(d_queue),contact->session);
if ((queued=creat(contact->mailname,S_IRUSR|S_IWUSR))<0)
  (void) clm_core_dump("%s, Unable to create <%s> (error=<%s>)",
		        OPOP,contact->mailname,strerror(errno));
if (reset==false) {
  (void) tracesid(contact,reset);
  (void) jrl_journalling(false,EVJRL,"%-*s[%s]/<%s>",6,"Peer:",
				     contact->peerip,contact->peername);
  if (strcmp(contact->servport,SMTPPORT)!=0)
    (void) jrl_journalling(false,EVJRL,"%-*s[%s:%s]",6,"Priv:",
				        contact->locip,contact->servport);
  }
(void) close(queued);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to update all "good" recipient*/
/*	In case of trouble after checking the	*/
/*	Email contents.				*/
/*						*/
/************************************************/
static void updctarcpt(RCPTYP **rcpts,char *info)

{
if (rcpts!=(RCPTYP **)0) {
  int i;

  for (i=0;rcpts[i]!=(RCPTYP *)0;i++) {
    register RCPTYP *rcptloc;

    rcptloc=rcpts[i];
    switch (rcptloc->status) {
      case m_indata	:
	rcptloc->why=clm_freestr(rcptloc->why);
	rcptloc->why=strdup(info);
	rcptloc->status=m_trouble;
	break;
      default		:	/*nothing to do	*/
	break;
      }
    }
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to convert "size=xxx" from	*/
/*	"MAIL FROM:" SMTP command, return	*/
/*	the maiel size value.			*/
/*						*/
/************************************************/
static int gettaille(char *buffer)

{
u_long mtaille;
char *ptr;

mtaille=0;
if ((ptr=strstr(buffer,"size"))!=(char *)0) {
  if ((ptr=strchr(ptr,'='))!=(char *)0) {
    ptr++;
    (void) sscanf(ptr,"%lu",&mtaille);
    }
  }
return mtaille;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare the header add on	*/
/*	received information			*/
/*						*/
/************************************************/
static void mkreceived(FILE *file,CTTYP *contact)

{
time_t curtime;
const char *smtp;
int phase;
int proceed;

curtime=time((time_t *)0);
smtp="SMTP";
phase=0;
proceed=true;
while (proceed) {
  switch (phase) {
    case 0	:
      (void) fprintf(file,"Received: from %s ([%s])\n",
			  contact->peername,contact->peerip);
      break;
    case 1	:
      if (contact->esmtp==true)
	smtp="ESMTP";
      (void) fprintf(file,"\tby %s ([%s:%s]/%s-%s) with %s\n",
                          contact->locname,contact->locip,contact->servport,
			  appname,clm_getvers(),smtp);
      break;
    case 2	:
      if (contact->cipherid!=(char *)0)
        (void) fprintf(file,"\t%s\n",contact->cipherid);
      break;
    case 3	:
      (void) fprintf(file,"\tid <%s@%s>",contact->session,contact->locname);
      break;
    case 4	:
      if (contact->goodrcpt==1) 
        (void) fprintf(file," for <%s>",contact->foremail);
      (void) fprintf(file,";\n");
      break;
    case 5	:
      (void) fprintf(file,"\t%s\n",clm_ascsysstamp(curtime));
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract and verify email	*/
/*	address within the SMTP protocol	*/
/*	Return the E-mail address within a	*/
/*	dynamically allocated memory, buffer	*/
/*	variable is updated with the remaining	*/
/*	data.					*/
/*						*/
/************************************************/
static char *extract(char *toextract,char *errstr,int taille,int from)

{
char *email;
char *buffer;
int proceed;
int phase;

email=(char *)0;
buffer=strdup(toextract);
proceed=true;
phase=1;
while (proceed==true) {
  char *ptr;

  ptr=(char *)0;
  switch (phase) {
    case   0	:	/*error collector	*/
      if (email!=(char *)0) {
	(void) free(email);
	email=(char *)0;
        }
      proceed=false;
      break;
    case   1	:
      if ((ptr=strchr(buffer,'<'))!=(char *)0) {
	(void) memmove(buffer,ptr+1,strlen(ptr+1)+1);
	}
      else {
	(void) snprintf(errstr,taille,"Syntax error, '<' not found");
	phase=-1;
	}
      break;
    case   2	:
      if ((ptr=strchr(buffer,'>'))!=(char *)0) {
	*ptr='\000';
	email=strdup(buffer);
	ptr++;
	(void) clm_skipblank(ptr);
        }
      else {
	(void) snprintf(errstr,taille,"Syntax error, '>' not found");
	phase=-1;
	}
      break;
    case   3	:
      if ((ptr=strchr(email,':'))!=(char *)0) 
	(void) memmove(email,ptr+1,strlen(ptr+1)+1);
      (void) clm_strtolower(email);
      break;
    case   4	:	/*empty from?				*/
      if ((strlen(email)==0)&&(from==true))
	phase=999;	/*empty email is OK for FROM:		*/
      break;
    case   5	:	/*postmaster				*/
      if ((from==false)&&(strcasecmp(email,POSTM)==0)) {
	phase=999;	/*<postmaster> is OK for TO:		*/
	}
      break;
    case   6	:
      if ((ptr=strchr(email,'@'))!=(char *)0) {
	if (strrchr(email,'@')!=ptr) {
	  (void) snprintf(errstr,taille,
			  "Syntax error, too many '@' charactere");
	  phase=-1;
          }
	}
      else {
	(void) snprintf(errstr,taille,
			"Syntax error, missing '@' charactere");
	phase=-1;
	}
      break;
    case   7	:	/*checking if there wrong char in email	*/
      if (email!=(char *)0) {
	char *ptre;

        ptre=email;
	while ((*ptre!='\000')&&(phase>0)) {
	  switch (*ptre) {
	    case ' '	:
	      (void) snprintf(errstr,taille,
			      "Syntax error, wrong char within Email");
	      phase=-1;
	      break;
	    default	:
	      break;
	    }
	  ptre++;
	  }
        }
      break;
#ifdef BUGRFC
    case   9	:	/*removing "'" type caractere		*/
      if (strchr(email,'\'')!=(char *)0)
	phase=-1;
      break;
#endif
    default	:	/*Everything fine			*/
      proceed=false;
      break;
    }
  phase++;
  }
buffer=clm_freestr(buffer);
return email;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check if an E-mail address	*/
/*	is correct.				*/
/*						*/
/************************************************/
static int isemailok(char *streml,char *errstr,int taille)

{
int status;

status=CMDOK;
if (strlen(streml)>0) {
  int proceed;
  int phase;
  char *localpart;
  char *domain;
  char *ptr;

  proceed=true;
  phase=1;
  localpart=strdup(streml);
  ptr=strchr(localpart,'@');
  domain=strdup(ptr+1);
  *ptr='\000';
  while (proceed==true) {
    switch (phase)  {
      case   0	:	/*error collector	*/
	status=BADPAR;
        proceed=false;
        break;
      case   1	:
	if (strlen(localpart)==0) {
	  (void) snprintf(errstr,taille,"Syntax error, missing localpart");
	  phase=-1;
	  }
	else {
	  int i;

	  for (i=0;(localpart[i]!='\000')&&(phase>0);i++) {
	    switch (localpart[i]) {
	      case ' '	:
	      case '<'	:
	      case '>'	:
	      case '('	:
	      case '['	:
	      case ']'	:
	      case ';'	:
	      case ':'	:
	        (void) snprintf(errstr,taille,"Syntax error, '%c' not allowed "
					      "in localpart",localpart[i]);
	        phase=-1;
		break;
	      default	:
		break;
	      }
	    }
	  }
        break;
      case   2	:	/*is domain existing?	*/
	if (strlen(domain)==0) {
	  (void) snprintf(errstr,taille,"Syntax error, missing domain part");
	  phase=-1;
	  }
        break;
      case   3	:	/*is domain ip address	*/
	if ((domain!=(char *)0)&&(domain[0]=='[')) {
	  AFNTYP *afnnum;

	  phase=999; 	/*checking IP address	*/
	  if ((afnnum=dns_isipaddress(domain))!=(AFNTYP *)0) 
	    afnnum=afn_freeipnum(afnnum);
	  else {
	    (void) snprintf(errstr,taille,"Syntax error, wrong domain part '%s' ",domain);
	    phase=-1;
	    }
	  }
	break;
      default	:
        proceed=false;
        break;
      }
    phase++;
    }
  (void) free(domain);
  (void) free(localpart);
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract authentication	*/
/*	info.					*/
/*	Return the login sequence		*/
/*						*/
/************************************************/
static char *getauth(CTTYP *contact,char *buffer)

{
#define	OPGA	"lvleml.c:getauth,"

static char *vocloc[]={"plain","login",(char *)0};
static char *logdat[]={"VXNlcm5hbWU6", /*Username: in B64	*/
		       "UGFzc3dvcmQ6", /*Password: in B64	*/
		       (char *)0};
char *login;
char *ptr;
int code;
int i;
char *auth;

login=(char *)0;
auth=strdup(buffer);
if ((ptr=strchr(buffer,' '))!=(char *)0) {
  *ptr='\000';
  (void) strcpy(auth,buffer);
  ptr++;
  }
else
  ptr="";
(void) memmove(buffer,ptr,strlen(ptr)+1);
for (code=0;vocloc[code]!=(char *)0;code++) {
  if (strcasecmp(auth,vocloc[code])==0) {
    break;
    }
  }
switch(code) {
    case  0  	:
      if (strlen(buffer)>0) {
        login=cnv_getb64(buffer);
        if ((login!=(char *)0)&&(strlen(login))>0) {
	  register int skip;
	  register char *lptr;

	  skip=strlen(IOBNULL)-1;
	  /*skipping authorization-id if present	*/
	  lptr=strstr(login,IOBNULL);	
	  /*keeping authentication-id and password	*/
	  if (lptr!=login) {
	    (void) memmove(login,lptr,strlen(lptr)+1);
	    lptr=login;
	    }
          while ((lptr=strstr(lptr,IOBNULL))!=(char *)0) {
  	    *lptr=' ';
  	    lptr++;
  	    (void) memmove(lptr,lptr+skip,strlen(lptr+skip)+1);
	    }
	  }
	}
      else {
	(void) clm_alert(0,"%s Missing authentication parameters",OPGA);
	(void) eml_report(contact,BADPAR,"Missing authentication data");
	}
      break;
    case  1  	:
      i=0;
      if (strlen(buffer)>0) {
        login=cnv_getb64(buffer);
	i=1;
	}
      for (;i<2;i++) {
        char *strloc;
	
	strloc=(char *)0;
	switch (i) {
	  case 0	:	/*need to send "username"	*/
            (void) eml_printfcrlf(contact,"%d %s",SENDB64,logdat[i]);
            if ((strloc=tls_waitforaction(contact->tls,contact->maxwait))!=(char *)0) {
              login=cnv_getb64(strloc);
	      (void) free(strloc);
	      }
	    break;
	  case 1	:	/*sending "password"		*/
            (void) eml_printfcrlf(contact,"%d %s",SENDB64,logdat[i]);
            if ((strloc=tls_waitforaction(contact->tls,contact->maxwait))!=(char *)0) {
	      char *prv;
	      char *clear;

	      prv=login;
	      clear=cnv_getb64(strloc);
	      (void) asprintf(&login,"%s %s",prv,clear);
	      (void) free(clear);
	      (void) free(prv);
	      (void) free(strloc);
	      }
	    break;
	  }
	}
      break;
    default  	:
      (void) clm_alert(0,"%s Unknown <%s> SMTP authentication mode",OPGA,auth);
      (void) eml_report(contact,BADAUTH,auth);
      break;
    }
(void) free(auth);
return login;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage a "MAIL FROM:" 	*/
/*	command from the SMTP client.		*/
/*						*/
/************************************************/
static char *checkfrom(CTTYP *contact,char *buffer)

{
char *answ;
char *mfrom;
char *efrom;
int phase;
int proceed;

answ=(char *)0;
mfrom=(char *)0;
efrom=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  char errstr[200];

  (void) strcpy(errstr,"");
  switch (phase) {
    case    0	:	/*previous from?		*/
      if (contact->frm!=(FRMTYP *)0) {
	(void) baddatareport(contact,MAILF,buffer,BADPAR,"Sender already specified");
	proceed=false;
	}
      break;
    case    1	:	/*checking for ':'		*/
      if (checkcolon(contact,buffer)==false)
	proceed=false;
      break;
    case    2	:	/*checking From format		*/
      if ((mfrom=extract(buffer,errstr,sizeof(errstr),true))==(char *)0) {
	(void) baddatareport(contact,MAILF,buffer,BADPAR,errstr);
	proceed=false;
        }
      break;
    case    3	:	/*from empty plus authentication*/
      efrom=strdup(mfrom);
      if ((strlen(mfrom)==0)&&(contact->autok==true)) {
	(void) free(efrom);
	efrom=strdup(contact->authname);
	phase++;	/*NO need to check special case	*/
	}
      break;
    case    4	:	/*checking for special case	*/
      if (strchr(mfrom,'@')==(char *)0) {
	char *newfrom;

	if (strlen(mfrom)==0) 
	  newfrom=strdup(POSTM);
	else
	  newfrom=strdup(mfrom);
	(void) free(efrom);
	if (contact->heloname!=(char *)0) 
	  (void) asprintf(&efrom,"%s@%s",newfrom,contact->heloname);
	else
	  (void) asprintf(&efrom,"%s@[%s]",newfrom,contact->peerip);
	(void) free(newfrom);
	}
      break;
    case    5	:	/*checking from format		*/
      if (isemailok(efrom,errstr,sizeof(errstr))!=CMDOK) {
	(void) baddatareport(contact,MAILF,buffer,BADPAR,errstr);
        phase=999;
        }
      break;
    case    6	:	/*everything is OK		*/
      contact->frm=(FRMTYP *)calloc(1,sizeof(FRMTYP));
      contact->frm->mfrom=strdup(mfrom);
      contact->frm->efrom=strdup(efrom);
      contact->frm->peerip=strdup(contact->peerip);
      contact->frm->locip=strdup(contact->locip);
      contact->frm->servport=strdup(contact->servport);
      (void) asprintf(&answ,"%s",buffer);
      break;
    case    7	:	/*extracting mail size		*/
      contact->frm->mtaille=gettaille(buffer);
      if (contact->badehlo==true) {
	if (strstr(contact->heloname,eml_getdomain(mfrom))==0) 
	  contact->frm->frmstatus|=BADEHLO;
	}
      break;
    default	:	/*Safe Guard, exiting		*/
      efrom=clm_freestr(efrom);
      mfrom=clm_freestr(mfrom);
      proceed=false;
      break;
    }
  phase++;
  }
return answ;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage a "RCPT TO:"	*/
/*	command from the SMTP client.		*/
/*						*/
/************************************************/
static char *checkto(CTTYP *contact,char *buffer)

{
char *to;
int phase;
int proceed;
char errstr[200];

to=(char *)0;
(void) strcpy(errstr,"");
phase=1;
proceed=true;
while (proceed==true) {
  int status;

  status=BADPAR;
  switch (phase) {
    case    0	:	/*phase out			*/
      to=clm_freestr(to);
      proceed=false;
      break;
    case    1	:	/*previous from?		*/
      if (contact->frm==(FRMTYP *)0) {
	(void) baddatareport(contact,RCPT,buffer,status,"Sender must be specified");
	phase=-1;
	}
      break;
    case    2	:	/*checking for ':'		*/
      if (strchr(buffer,':')==(char *)0) {
	(void) baddatareport(contact,RCPT,buffer,SYNERR,"':' expected");
	phase=-1;
        }
      break;
    case    3	:	/*extracting address format	*/
      if ((to=extract(buffer,errstr,sizeof(errstr),false))==(char *)0) {
	(void) baddatareport(contact,RCPT,buffer,status,errstr);
        phase=-1;
        }
      break;
    case    4	:	/*could the address an empty one	*/
      if (strlen(to)==0) {
	(void) baddatareport(contact,RCPT,buffer,status,"Recipient must be specified");
	phase=-1;
	}
      break;
    case    5	:	/*could be postmaster?			*/
      if (strcmp(to,POSTM)==0)	/*yes no need to go further	*/
	phase=999;
      break;
    case    6	:	/*checking adresse format		*/
      if ((status=isemailok(to,errstr,sizeof(errstr)))!=CMDOK) {
	(void) baddatareport(contact,RCPT,buffer,status,errstr);
        phase=-1;
        }
      break;
    default	:	/*Safe Guard, everything fine	*/
      proceed=false;
      break;
    }
  phase++;
  }
return to;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to return the keyword list	*/
/*	according the protocol used.		*/
/*						*/
/************************************************/
static const VOCTYP *getkwlst(PROTYP proto)

{
const VOCTYP *list;

list=vocsmtp;
switch(proto) {
  case pro_smtps	:
  case pro_smtp		:
    break;
  case pro_pop3		:
    list=(VOCTYP *)0;
    (void) clm_alert(0,"lvleml.c:getkwlst, proto pop3 to be defined");
    break;
  default		:
    (void) clm_alert(0,"lvleml.c:getkwlst, Unexpected prototyp '%d'",proto);
  }
return list;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send an HELP explaination	*/
/*	to remote server.			*/
/*						*/
/************************************************/
static void dohelp(CTTYP *cta,const VOCTYP *kwlst,char *cmd)


{
#define	HLPMRK	214

if (strlen(cmd)>0) {
  VOCTYP *voc;

  if ((voc=clm_getvoca(kwlst,cmd))!=(VOCTYP *)0) {
    ACTTYP *acting;

    acting=(ACTTYP *)(voc->ptr);
    if ((acting!=(ACTTYP *)0)&&(acting->help!=(char **)0)) {
      int i;

      for (i=0;acting->help[i]!=(char *)0;i++) {
        (void) eml_printfcrlf(cta,"%d- %s",HLPMRK,acting->help[i]);
        }
      }
    else {
      (void) eml_printfcrlf(cta,"%d HELP not available",HLPMRK);
      }
    (void) eml_printfcrlf(cta,"%d  End of HELP info",HLPMRK);
    } 
  else {
    (void) eml_printfcrlf(cta,"%d HELP topic \"%s\" unknown",HLPMRK,cmd);
    }
  }
else {
  int i;
  int cpt;
  char helpstr[200];

  (void) eml_printfcrlf(cta,"%d- This is %s version %s",HLPMRK,
			  appname,clm_getvers());
  (void) eml_printfcrlf(cta,"%d- Topics",HLPMRK);
  for (i=0,cpt=0;kwlst[i].key!=(char *)0;i++) {
    ACTTYP *acting;

    acting=(ACTTYP *)kwlst[i].ptr;
    if (acting->help!=(char **)0) {
      char dhelp[200];
      char help[200];

      if ((cpt%5)==0)
        (void) strcpy(helpstr,"");
      (void) snprintf(dhelp,sizeof(dhelp),"'%s'",kwlst[i].key);
      (void) snprintf(help,sizeof(help),"%-14s",dhelp);
      (void) strcat(helpstr,help);
      if ((cpt%5)==4) {
        (void) eml_printfcrlf(cta,"%d- %s",HLPMRK,helpstr);
        (void) strcpy(helpstr,"");
	}
      cpt++;
      }
    }
  if (strlen(helpstr)>0)
    (void) eml_printfcrlf(cta,"%d- %s",HLPMRK,helpstr);
  (void) eml_printfcrlf(cta,"%d- For more info use \"HELP <topic>\".",HLPMRK);
  (void) eml_printfcrlf(cta,"%d- To report bugs in the implementation "
			  "send email to",HLPMRK);
  (void) eml_printfcrlf(cta,"%d-\t%s@%s",HLPMRK,appname,SAFEDOM);
  (void) eml_printfcrlf(cta,"%d- For local information send email to "
			  "Postmaster at your site.",HLPMRK);
  (void) eml_printfcrlf(cta,"%d  End of HELP info",HLPMRK);
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to see if we can collect	*/
/*	E-mail data part.			*/
/*	Return a trouble status			*/
/*						*/
/************************************************/
static SCODTYP godata(CTTYP *cta)

{
SCODTYP scode;
int phase;
int proceed;

scode=s_baddata;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case  0	:	/*check if we know from	*/
      if (cta->frm==(FRMTYP *)0) {
	(void) eml_printfcrlf(cta,"%d Need MAIL command",BADSEQ);
	phase=999;	/*get out fast		*/
	}
      break;
    case  1	:	/*at least one RCPT?	*/
      if (countrcpts(cta->rcpts)==0) {
	(void) eml_printfcrlf(cta,"%d Need RCPT (recipient)",BADSEQ);
	phase=999;	/*get out fast		*/
	}
      break;
    case  2	:	/*seems to be good 	*/
      scode=s_data;	/*lets try to collect	*/
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return scode;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect email header	*/
/*	received from remote SMTP site.		*/
/*						*/
/************************************************/
static int collectheader(CTTYP *cta)

{
#define	OPCO	"lvleml.c:collectheader"

#define	HDRLIM	65536	/*max header size	*/
#define	TOOBIG 	"Abort! E-mail header too big (current %lu / maxsize=%d)"

int headerok;
FILE *store;
char *hdrname;
int phase;
int proceed;

headerok=false;
store=(FILE *)0;
(void) asprintf(&hdrname,"%s%s",cta->mailname,strpart[HDPART]);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening file		*/
      if ((store=fopen(hdrname,"w"))==(FILE *)0) {
        char *trouble;

        trouble=eml_ascstat(ERRPROC,"Unable to open file");
        (void) eml_printfcrlf(cta,"%s <%s>",trouble,strerror(errno));
        (void) jrl_journalling(false,EVJRL,"%s <%s> (errno=<%s>, bug?)",
				     trouble,hdrname,strerror(errno));
  	(void) free(trouble);
  	(void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",OPCO,
		            hdrname,strerror(errno));
	proceed=false;	/*big big trouble!	*/
	}
      break;
    case 1	:	/*preparing email header*/
      (void) mkreceived(store,cta);
      (void) eml_report(cta,DATAOK,"");
      break;
    case 2	:	/*collecting header	*/
      while (proceed==true) {
        char *data;

	if ((data=tls_waitforaction(cta->tls,cta->maxwait))==(char *)0) {
	  phase=999;	/*trouble trouble	*/
	  break;	/*no need to go further	*/
	  }
        data=clm_skiprtlf(data);
	if (strcmp(data,DATAEND)==0) {
          (void) jrl_journalling(false,EVJRL,"Received DATAEND within header!");
	  break;
	  }
	if (data[0]=='\000') {	/*header end	*/
	  (void) free(data);	
	  headerok=true;
	  break;
	  }
	if (strstr(data,XCLEM)==data) {
	  (void) free(data);	/*removing XCLEM*/
	  continue;
	  }
        cta->mtaille+=strlen(data);
	cta->headers=(char **)vec_addlstlst((void **)cta->headers,
                                              (void *)strdup(data));
        (void) fprintf(store,"%s\n",data);
        (void) free(data);
	if (cta->mtaille>HDRLIM) {
	  char *info;

          (void) asprintf(&info,"Mail header too large (current %lu / max %d)",
                                 cta->mtaille,HDRLIM);
          (void) updctarcpt(cta->rcpts,info);
          (void) eml_report(cta,MALABRT,info);
          (void) jrl_journalling(false,EVJRL,TOOBIG,cta->mtaille,HDRLIM);
          cta->headers=(char **)vec_freelstlst((void **)cta->headers,VECFREE);
          (void) free(info);
	  break;	/*no need to go further	*/
	  }
	}
      break;
    default	:	/*SAFE guard		*/
      if (fclose(store)<0) {
        char *trouble;

        trouble=eml_ascstat(ERRPROC,"Unable to close file");
        (void) eml_printfcrlf(cta,"%s <%s>",trouble,strerror(errno));
        (void) jrl_journalling(false,EVJRL,"%s <%s> (errno=<%s>, bug?)",
				     trouble,hdrname,strerror(errno));
  	(void) free(trouble);
  	(void) clm_alert(0,"%s Unable to close <%s> (error=<%s>)",OPCO,
		            hdrname,strerror(errno));
	headerok=false;
	}
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(hdrname);
return headerok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect email parts	*/
/*	received from remote SMTP site.		*/
/*						*/
/************************************************/
static int collectmain(CTTYP *cta)

{
/*absolute email limit				*/
#define	MAILIM	(1024*1024*1024)	/*1Gig!	*/

int mainok;
FILE *store;
char *dataname;
int phase;
int proceed;

mainok=false;
store=(FILE *)0;
(void) asprintf(&dataname,"%s%s",cta->mailname,strpart[DTPART]);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening file		*/
      if ((store=fopen(dataname,"w"))==(FILE *)0) {
        char *trouble;

        trouble=eml_ascstat(ERRPROC,"Unable to open file");
        (void) eml_printfcrlf(cta,"%s <%s>",trouble,strerror(errno));
        (void) jrl_journalling(false,EVJRL,"%s <%s> (errno=<%s>, bug?)",
				     trouble,dataname,strerror(errno));
  	(void) free(trouble);
  	(void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",OPCO,
		            dataname,strerror(errno));
	proceed=false;	/*big big trouble!	*/
	}
      break;
    case 1	:	/*preparing email header*/
      (void) fprintf(store,"\n");/*empty line	*/
      while (proceed==true) {
        char *data;

	if ((data=tls_waitforaction(cta->tls,cta->maxwait))==(char *)0) {
	  phase=999;	/*trouble trouble	*/
	  break;	/*no need to go further	*/
	  }
        cta->mtaille+=strlen(data);
        data=clm_skiprtlf(data);
	if (strcmp(data,DATAEND)==0) {	
	  mainok=true;
	  (void) free(data);
	  break;
	  }
	if ((data[0]=='.')&&(data[1]=='.'))
	  (void) memmove(data,data+1,strlen(data+1)+1);
	(void) fprintf(store,"%s\n",data);
	(void) free(data);
	}
      break;
    default	:	/*SAFE guard		*/
      if (fclose(store)<0) {
        char *trouble;

        trouble=eml_ascstat(ERRPROC,"Unable to close file");
        (void) eml_printfcrlf(cta,"%s <%s>",trouble,strerror(errno));
        (void) jrl_journalling(false,EVJRL,"%s <%s> (errno=<%s>, bug?)",
				     trouble,dataname,strerror(errno));
  	(void) free(trouble);
  	(void) clm_alert(0,"%s Unable to clode <%s> (error=<%s>)",OPCO,
		            dataname,strerror(errno));
	mainok=false;
	}
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(dataname);
return mainok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to search a string sequence	*/
/*	within a string array			*/
/*						*/
/************************************************/
static int ihaveinimap(char **imapdata,char *seq)

{
int found;

found=false;
if (imapdata!=(char **)0) {
  int i;

  for (i=0;imapdata[i]!=(char *)0;i++) {
    if (strstr(imapdata[i],seq)!=(char *)0) {
      found=true;
      break;
      }
    }
  }
return found;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect the full imap 	*/
/*	answer from remot server		*/
/*						*/
/************************************************/
static int getimapreply(TLSFILE *tls,int maxtime,const char *seq,char ***reply)

{
int status;
int found;
char *buffer;

status=REJECT;
found=false;
buffer=(char *)0;
while (tls_getreply(tls,maxtime,pro_imap,&buffer)==CMDOK) {
  *reply=(char **)vec_addlstlst((void **)*reply,(void *)buffer);
  if (strstr(buffer,seq)==buffer) {
    char *ptr;

    ptr=buffer+strlen(seq)+1;
    if (strstr(ptr,"OK")==ptr) {
      status=CMDOK;
      found=true;
      }
    break;
    }
  buffer=(char *)0;
  }
if (found==false)	/*sequence not found!	*/
  *reply=(char **)vec_freelstlst((void **)*reply,VECFREE);
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to request current capability	*/
/*	to an imap server.			*/
/*						*/
/************************************************/
static char **imapcapability(TLSFILE *tls,char *rmtauth,char *authport,const char *seq)


{
#define	OPCA	"lvleml.c:imapcapability,"

char **reply;
int phase;
int proceed;

reply=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*requesting capability	*/
      if (tls!=(TLSFILE *)0) {	/*always	*/
	int taille;
	char *buffer;

        taille=asprintf(&buffer,"%s %s%s",seq,"CAPABILITY",CRLF);
        if (tls_write(tls,buffer,taille)!=taille) {
	  (void) clm_alert(0,"%s Unable to request capabilty to IMAP peer <%s:%s>",OPCA,rmtauth,authport);
	  phase=999;	/*get out fast		*/
	  }
        buffer=clm_freestr(buffer);
        }
      break;
    case 1	:	/*got capability?	*/
      if (getimapreply(tls,MXTIME,seq,&reply)!=CMDOK) {
	(void) clm_alert(0,"%s Unable to get answer to capabilty from IMAP peer <%s:%s>",OPCA,rmtauth,authport);
	phase=999;	/*get out fast		*/
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return reply;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to override remote IP number	*/
/*	with a new IP number.			*/ 
/*						*/
/************************************************/
void eml_doorgn(CTTYP *cta,char *nip)

{
int phase;
int proceed;

phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case   0	:	/*error trap		*/
      proceed=false;
      break;
    case   1	:	/*checking good ipnum	*/
      if (strlen(nip)==0) {
	(void) eml_report(cta,BADPAR,"IP number missing");
	phase=-1;
        }
      break;
    case   2	: 	/*getting host		*/
      if (strlen(nip)>0) {	/*always	*/
	int status;
	struct addrinfo hints;
	struct addrinfo *res;
	
	(void) memset(&hints,'\000',sizeof(hints));
	hints.ai_family=PF_UNSPEC;
	hints.ai_socktype=SOCK_STREAM;
	hints.ai_flags=HINTFLG;
	status=getaddrinfo(nip,"",&hints,&res);
	if (status!=0) {
	  (void) eml_report(cta,BADPAR,"Not an IP number");
	  phase=-1;
	  }
	else {
	  int namestat;
	  char host[NI_MAXHOST] = NORVERS;
	  char *pnip;	

	  namestat=getnameinfo(res->ai_addr,res->ai_addrlen,host,sizeof(host),(char *)0,0,NI_NAMEREQD);
	  pnip=nip;
	  if (res->ai_family==AF_INET6) 
	    (void) asprintf(&pnip,"%s%s",STRIPV6,nip);
	  switch (namestat) {
	    case 0		:
	    case EAI_AGAIN	:
	    case EAI_NONAME	:
	      break;
	    default		:
	      (void) clm_alert(0,"%s Unexpected namestat='%d' trying to resolv <%s> IP number reverse-address (bug?)",
				 "lvleml.c:eml_doorgn,",namestat,pnip);
	      (void) eml_report(cta,BADPAR,"Not an IP number");
	      phase=-1;
	      break;
	    }
	  (void) free(cta->peername);
	  (void) free(cta->peerip);
	  cta->peername=strdup(host);
	  cta->peerip=strdup(pnip);
	  if (pnip!=nip)
	    (void) free(pnip);
	  }
	}
      break;
    case   3	:	/*setting contact value	*/
      break;
    case   4	:	/*updating process info	*/
      (void) emlsettitle(cta);
      (void) eml_report(cta,CMDOK,"");
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to move a pop-3 link in tLS	*/
/*	mode.					*/
/*						*/
/************************************************/
int eml_setpop3tls(TLSFILE *tls)

{
#define	OPPT	"lvleml.c:eml_setpop3tls,"

#define	STLS	"STLS"

int tlsdone;
int phase;
int proceed;

tlsdone=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sending username	*/
      if (tls!=(TLSFILE *)0) {	/*always	*/
	char *buffer;
	int taille;

	buffer=(char *)0;	
	taille=asprintf(&buffer,"%s%s",STLS,CRLF);
	if (tls_write(tls,buffer,taille)!=taille) {
	  (void) clm_alert(0,"%s Unable to send TLS command <%s>",OPPT,buffer);
	  phase=999;	/*get out fast		*/
	  }
        buffer=clm_freestr(buffer);
	}
      break;
    case 1	:	/*expecting command ack	*/
      if (tls!=(TLSFILE *)0) {	/*always	*/
	char *buffer;

	buffer=(char *)0;	
        if (tls_getreply(tls,MXTIME,pro_pop3t,&buffer)!=CMDOK) {
	  (void) clm_alert(0,"%s Unable to get answer once <%s> request sent",OPPT,STLS);
	  phase=999;	/*get out fast		*/
	  }
        buffer=clm_freestr(buffer);
	}
      break;
    case 2	:	/*establishing TLS mode	*/
      if (tls_gotls(tls,false)<0) {
	(void) clm_alert(0,"%s Unable to negociate TLS with pop3 site)",OPPT);
	phase=999;	/*Big trouble		*/
	}
      break;
    case 3	:	/*successful		*/
      tlsdone=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return tlsdone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to move a pop-3 link in tLS	*/
/*	mode.					*/
/*						*/
/************************************************/
int eml_setimaptls(TLSFILE *tls)

{
#define	OPAP	"lvleml.c:eml_setimaptls,"

int tlsdone;
int phase;
int proceed;

tlsdone=false;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sending username	*/
      if (tls!=(TLSFILE *)0) {	/*always	*/
	const char *seq;
	char *buffer;
	int taille;

	seq=getimapseq(phase);
	buffer=(char *)0;	
	taille=asprintf(&buffer,"%s %s%s",seq,STRTTLS,CRLF);
	if (tls_write(tls,buffer,taille)!=taille) {
	  (void) clm_alert(0,"%s Unable to send TLS command <%s>",OPAP,buffer);
	  phase=999;	/*get out fast		*/
	  }
        buffer=clm_freestr(buffer);
	}
      break;
    case 1	:	/*expecting command ack	*/
      if (tls!=(TLSFILE *)0) {	/*always	*/
	const char *seq;
	char **buffer;

	seq=getimapseq(phase-1);
	buffer=(char **)0;	
        if (getimapreply(tls,MXTIME,seq,&buffer)!=CMDOK) {
	  (void) clm_alert(0,"%s Unable to get answer once <%s> request sent",OPAP,STRTTLS);
	  phase=999;	/*get out fast		*/
	  }
        buffer=(char **)vec_freelstlst((void **)buffer,VECFREE);
	}
      break;
    case 2	:	/*establishing TLS mode	*/
      if (tls_gotls(tls,false)<0) {
	(void) clm_alert(0,"%s Unable to negociate TLS with imap site)",OPAP);
	phase=999;	/*Big trouble		*/
	}
      break;
    case 3	:	/*successful		*/
      tlsdone=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return tlsdone;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a user password via	*/
/*	POP3 Protocol.				*/
/*						*/
/************************************************/
int eml_checkviapop3(TLSFILE *tls,char *username,char *password)

{
#define	OPCP	"lvleml.c:lvl_checkviapop3,"

int goodid;
char *buffer;
int taille;
int phase;
int proceed;

goodid=false;
buffer=(char *)0;
taille=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sending username	*/
      taille=asprintf(&buffer,"%s %s%s","USER",username,CRLF);
      if (tls_write(tls,buffer,taille)!=taille) {
	(void) clm_alert(0,"%s Unable to send user name <%s>",OPCP,username);
	phase=999;	/*get out fast		*/
	}
      buffer=clm_freestr(buffer);
      break;
    case 1	:	/*expecting command ack	*/
      buffer=(char *)0;
      if (tls_getreply(tls,MXTIME,pro_pop3,&buffer)!=CMDOK) {
	(void) clm_alert(0,"%s Unable to get answer once user name <%s> sent",OPCP,username);
	phase=999;	/*get out fast		*/
	}
      buffer=clm_freestr(buffer);
      break;
    case 2	:	/*sending password	*/
      taille=asprintf(&buffer,"%s %s%s","PASS",password,CRLF);
      if (tls_write(tls,buffer,taille)!=taille) {
	(void) clm_alert(0,"%s Unable to send password <%s> for user <%s>",
			    OPCP,password,username);
	phase=999;	/*get out fast		*/
	}
      buffer=clm_freestr(buffer);
      break;
    case 3	:	/*expecting answer	*/
      if (tls_getreply(tls,MXTIME,pro_pop3,&buffer)==CMDOK) {
	goodid=true;
	}
      buffer=clm_freestr(buffer);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return goodid;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to check a user password via	*/
/*	IAMP Protocol.				*/
/*						*/
/************************************************/
int eml_checkviaimap(TLSFILE *tls,char *rmtauth,char *authport,char *username,char *password)

{
#define	OPIP	"lvleml.c:lvl_checkviaiamp,"

#define	APLAIN	"AUTH=PLAIN"

int goodid;
char **reply;
int phase;
int proceed;

goodid=false;
reply=(char **)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*capabilities		*/
      if ((reply=imapcapability(tls,rmtauth,authport,getimapseq(phase)))==(char **)0)
	phase=999;
      break;
    case 1	:	/*AUTH=PLAIN avail?	*/
      if (ihaveinimap(reply,APLAIN)==false) {
	(void) clm_alert(0,"%s, <%s> authentication chanel is not available with IMAP server  <%s:%s>",OPIP,APLAIN,rmtauth,authport);
	phase=999;
	}
      reply=(char **)vec_freelstlst((void **)reply,VECFREE);
      break;
    case 2	:	/*sending authentication*/
      if (tls!=(TLSFILE *)0) {	/*always	*/
	const char *seq;
	char *auth;
	int taille;
	char *buffer;

	seq=getimapseq(phase);
	(void) asprintf(&buffer,"\\000%s\\000%s",username,password);
	auth=cnv_setb64(buffer);
	(void) free(buffer);
        taille=asprintf(&buffer,"%s authenticate plain %s%s",seq,auth,CRLF);
	(void) free(auth);
        if (tls_write(tls,buffer,taille)!=taille) {
	  (void) clm_alert(0,"%s Unable to request capability to IMAP peer <%s:%s>",OPCA,rmtauth,authport);
	  phase=999;	/*get out fast		*/
	  }
        buffer=clm_freestr(buffer);
        }
      break;
    case 3	:	/*authentication result	*/
      if (getimapreply(tls,MXTIME,getimapseq(phase-1),&reply)==CMDOK) 
	goodid=true;
      reply=(char **)vec_freelstlst((void **)reply,VECFREE);
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return goodid;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to open a session log		*/
/*						*/
/************************************************/
void eml_opensession(CTTYP *contact)

{
(void) opensession(contact,false);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reset session structure	*/
/*	to be ready for a new E-mail.		*/
/*						*/
/************************************************/
void eml_resetsession(CTTYP *contact)

{
(void) closesession(contact);
contact->reset=true;
(void) opensession(contact,contact->reset);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to reset contact structure	*/
/*	to be ready for a new E-mail.		*/
/*						*/
/************************************************/
void eml_closesession(CTTYP *contact)

{
(void) closesession(contact);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to handle exchange with remote*/
/*	SMTP client, if the command can't be	*/
/*	handle in this routine, information is	*/
/*	sent back to the calling routine.	*/
/*						*/
/************************************************/
SCODTYP eml_docontact(CTTYP *cta)

{
#define	OPDM	"lvleml.c:eml_docontact,"

SCODTYP scode;
const VOCTYP *kwlst;
int proceed;

scode=s_unknown;
kwlst=getkwlst(cta->proto);
proceed=true;
while (proceed==true) {
  VOCTYP *voc;
  ACTTYP *acting;
  char *buffer;

  voc=(VOCTYP *)0;
  acting=(ACTTYP *)0;
  scode=s_unknown;
  if ((buffer=tls_waitforaction(cta->tls,cta->maxwait))==(char *)0) {
    scode=s_timeout;
    (void) jrl_journalling(false,EVJRL,"Session Time-Out");
    break;
    }
  (void) clm_cleanstring(buffer);
  voc=clm_getvoca(kwlst,buffer);
  if ((voc==(VOCTYP *)0)||(voc->code==c_unknown)) {
    (void) synerr(cta,buffer);
    buffer=clm_freestr(buffer);
    continue;
    }
  if ((acting=(ACTTYP *)voc->ptr)==(ACTTYP *)0) {
    (void) clm_core_dump("%s VOC->%d, without action (bug!)",OPDM,voc->code);
    }
  if (cta->reset==true) {
    cta->reset=false;
    if (voc->code!=c_quit) {
      (void) tracesid(cta,true);
      }
    }
  switch (voc->code) {
    case c_auth 	:
    case c_chkid	:       /*DO NOT standardize            */
      break;
    default     	:
      (void) eml_standardize(buffer);
      break;
    }
  scode=acting->scode;
  proceed=acting->proceed;
  switch (voc->code) {
    case c_help 	:
      (void) dohelp(cta,kwlst,buffer);
      break;
    case c_noop 	:
      (void) eml_printfcrlf(cta,"%d OK",CMDOK);
      break;
    case c_trusting  :
      proceed=false;
      break;
    case c_tls  :
      (void) tls_fpurge(cta->tls);/*CVE-2011-0411, "plaintext command injection"*/
      if (tls_intls(cta->tls)==false) {
        (void) eml_printfcrlf(cta,"%d Ready to start TLS",SIGNON);
        if (tls_gotls(cta->tls,true)==0) {
          (void) jrl_dbgjournalling(1,false,EVJRL,"ComLink set in TLS mode");
          (void) jrl_journalling(false,EVJRL,"Link now in crypted mode");
          cta->cipherid=tls_getcipher(cta->tls);
	  }
	}
      else {
        (void) eml_printfcrlf(cta,"%d Already in crypted mode",BADSEQ);
	}
      break;
    case c_orgn         :
      if (eml_orgn==true) {
        (void) eml_doorgn(cta,buffer);
        (void) jrl_journalling(false,EVJRL,"%s: [%s]/<%s>","Peer override",
				           cta->peerip,cta->peername);
	}
      else
        (void) eml_printfcrlf(cta,"%d Command forbiden",SYNERR);
      break;
    case c_mail 	:	/*checking from info	*/
      if (cta->reset==true) {
	cta->reset=false;
        (void) tracesid(cta,true);
	}
      cta->info=clm_freestr(cta->info);
      if ((cta->info=checkfrom(cta,buffer))==(char *)0)
	proceed=true;
      break;
    case c_rcpt 	:	/*checking recipient	*/
      cta->info=clm_freestr(cta->info);
      if ((cta->info=checkto(cta,buffer))==(char *)0)
	proceed=true;
      break;
    case c_reset 	:	/*reseting the SMTP link*/
      (void) jrl_journalling(false,EVJRL,"Main Session <%s> Reset ",cta->msession);
      break;			/*done at upper level	*/
    case c_ehlo		:
    case c_helo		:
      cta->info=clm_freestr(cta->info);
      cta->info=buffer;
      buffer=(char *)0;
      proceed=false;
      break;
    case c_data		:	/*collecting E-mail data*/
      proceed=false;
      scode=godata(cta);
      break;
    case c_quit		:	/*nothing to do anymore */
      (void) eml_report(cta,QUITOK,cta->locname);
      (void) jrl_journalling(false,EVJRL,"Session <%s> end",cta->msession);
      break;			/*getting out		*/
    case c_auth		:
      if (cta->autok==true) {
        (void) sleep(1);
	(void) eml_report(cta,BADSEQ,"Already Authenticated");
	proceed=true;
        }
      else {
        cta->info=clm_freestr(cta->info);
        if ((cta->info=getauth(cta,buffer))==(char *)0) {
	  proceed=true;
	  }
        }
      break;
    default		:
      (void) clm_alert(0,"%s VOC->%d, can't reach (bug!)",OPDM,voc->code);
      break;
    }
  buffer=clm_freestr(buffer);
  }
return scode;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a message in IPC	*/
/*	channel.				*/
/*	Procedure is waiting for remote to	*/
/*	answer and suppose there is only one	*/
/*	message from destination back at any	*/
/*	given time.				*/
/*	if trouble, return null			*/
/*						*/
/************************************************/
char *eml_xchgmsg(BPTR *bptr,u_long pool,char *fmt,...)

{
#define	OPXC	"lvleml.c:eml_xchgmsg,"

#ifdef DEBUG
#define	MXWMSG	150	/*wait 20 sec		*/
#else
#define	MXWMSG	150	/*wait 150 sec		*/
#endif

char *got;
va_list args;
char *msg;
char *msgref;
TIMESPEC timeout;
int phase;
int proceed;

got=(char *)0;
va_start(args,fmt);
(void) vasprintf(&msg,fmt,args);
msgref=(char *)0;
timeout.tv_sec=(time_t)MXWMSG;
timeout.tv_nsec=(u_long)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*sending message	*/
      if ((msgref=ipc_postmsg(bptr,&timeout,pool,msg))==(char *)0) {
	(void) clm_alert(0,"%s Unable to post message <%s> within '%d' sec",OPXC,msg,MXWMSG);
	phase=999;	/*trouble trouble!	*/
	}
      break;
    case 1	:	/*waiting for reply	*/
      timeout.tv_sec=(time_t)MXWMSG;
      timeout.tv_nsec=(u_long)0;
      if ((got=ipc_getreply(bptr,&timeout,msgref))!=(char *)0)
	phase=999;	/*everything fine!	*/
      break;
    case 2	:	/*so we have trouble	*/
      (void) clm_alert(0,"%s no answer to msg=<%s> within '%d' sec",OPXC,msg,MXWMSG);
      timeout.tv_sec=(time_t)MXWMSG;
      timeout.tv_nsec=(u_long)0;
      if (ipc_rmmymsg(bptr,&timeout,msgref)==true) {
        (void) clm_alert(0,"%s removed pending msg=<%s> from queue",OPXC,msg);
	phase=999;	/*lets fortget about it	*/
	}
      break;
    case 3	:	/*we are in real trouble*/
      timeout.tv_sec=(time_t)MXWMSG*2;
      timeout.tv_nsec=(u_long)0;
      (void) clm_alert(0,"%s waiting for '%d' sec more to msg=<%s> answer",OPXC,timeout.tv_sec,msg);
      if ((got=ipc_getreply(bptr,&timeout,msgref))!=(char *)0)
	phase=999;	/*msg received at last!	*/
      break;
    case 4	:	/*Message lost		*/
      (void) clm_alert(0,"%s Never got an answer to post <%s>, message LOST!",OPXC,msg);
      break;
    default	:	/*SAFE Guard		*/
      msgref=clm_freestr(msgref);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(msg);
va_end(args);
return got;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a message in the IPC	*/
/*	channel.				*/
/*	Procedure is NOT waiting for a reply	*/
/*						*/
/************************************************/
int eml_sendmsg(BPTR *bptr,u_long pool,time_t mxtime,char *fmt,...)

{
#define	OPSE	"lvleml.c:eml_sendmsg,"

int done;
va_list args;
char *msgref;
char *msg;
TIMESPEC timeout;

done=false;
va_start(args,fmt);
timeout.tv_sec=mxtime;
timeout.tv_nsec=(u_long)0;
(void) vasprintf(&msg,fmt,args);
if ((msgref=ipc_postmsg(bptr,&timeout,pool,msg))!=(char *)0) {
  (void) free(msgref);
  done=true;
  }
else
  (void) clm_alert(0,"%s Unable to post msg=<%s>",OPSE,msg);
(void) free(msg);
va_end(args);
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to report a status/reply to	*/
/*	an SMTP remote end.			*/
/*						*/
/************************************************/
void eml_report(CTTYP *contact,int status,char *cmd)

{
char *info;

info=eml_ascstat(status,cmd);
(void) eml_printfcrlf(contact,info);
(void) free(info);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove the distpache file	*/
/*						*/
/************************************************/
void eml_remdispatch(CTTYP *cta)

{
char *filename;

(void) asprintf(&filename,"%s-%s",cta->mailname,RCPTEXT);
if (unlink(filename)<0)
  (void) clm_alert(0,"%s Unable to remove file <%s> (error=<%s>)",
		     "lvleml.c:eml_remdispatch,",filename,strerror(errno));
(void) free(filename);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to manage a "HELO" commande	*/
/*	from the SMTP client.			*/
/*						*/
/************************************************/
int eml_checkehlo(CTTYP *cta,char *buffer,char sep)

{
int good;
char *helstr;
int phase;
int proceed;

good=false;
helstr="HELO";
if (sep=='-')
  helstr="EHLO";
cta->info=clm_freestr(cta->info);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*domain part missing?	*/
      if (strlen(buffer)==0) {
        cta->info=eml_ascstat(BADPAR,"domain part missing");
	(void) eml_printfcrlf(cta,"%s, disconnecting",cta->info);
        (void) jrl_journalling(false,EVJRL,"%s: %s\n\t  %s, disconnecting",
			       helstr,buffer,cta->info);
	phase=999;	/*trouble		*/
	}
      break;
    case 1	:	/*cleaning hello data	*/
      (void) clm_skipblank(buffer);
      break;
    case 2	:	/*is is a real domain/IP*/
      if (dns_isfqdn(buffer)==false) {
	cta->badehlo=true;
	}
      break;
    case 3	:	/*everything fine	*/
      cta->heloname=clm_freestr(cta->heloname);
      cta->heloname=strdup(buffer);
      if (cta->badehlo==false)
        (void) asprintf(&(cta->info),"%d%c%s Hello %s [%s], please to meet you",
			             CMDOK,sep,cta->locname,cta->peername,
			             cta->peerip);
      else
        (void) asprintf(&(cta->info),"%d%c%s Hum! %s [%s], No MX on your <%s> hello",
			             CMDOK,sep,cta->locname,cta->peername,
				     cta->peerip,cta->heloname);
      (void) eml_printfcrlf(cta,"%s",cta->info);
      good=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return good;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to create E-mail files (data	*/
/*	and queue) to be set ready in mail queue*/
/*						*/
/************************************************/
int eml_dispatch(CTTYP *contact)

{
#define	OPDI	"lvleml.c:eml_dispatch,"

int completion;
char *filename;
FILE *fichier;
DSPTYP *dsp;
int numrcpt;
int phase;
int proceed;

completion=true;
filename=(char *)0;
fichier=(FILE *)0;
dsp=(DSPTYP *)0;
numrcpt=0;
phase=1;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*error collector	*/
      completion=false;
      proceed=false;
      break;
    case 1	:	/*Check RCPT to relay	*/
      if (contact->rcpts!=(RCPTYP **)0) {
	int i;

	for (i=0;contact->rcpts[i]!=(RCPTYP *)0;i++) {
	  switch (contact->rcpts[i]->status) {
	    case m_trouble   :	/*can't comply	*/
	    case m_rejected  :	/*reject entry	*/
	    case m_noaddress :	/*void entry	*/
	      break;
	    default	     :
	      numrcpt++;
	      break;
	    }
	  }
	}
      if (numrcpt==0) {
        (void) clm_alert(0,"%s No recipient to dispatch session=<%s>",
			    OPDI,contact->session);
        phase=-1;	/*trouble?!(unpossible!)*/
 	}
      break;
    case 2	:	/*Generating DSP struct	*/
      if ((dsp=eml_contacttodsp(contact))==(DSPTYP *)0) {
        (void) clm_alert(0,"%s No dispatch struct (bug?)",OPDI);
        phase=-1;	/*trouble?!(unpossible!)*/
 	}
      break;
    case 3	:	/*generating Q-file	*/
      (void) asprintf(&filename,"%s-%s",contact->mailname,RCPTEXT);
      if ((fichier=fopen(filename,"w"))!=(FILE *)0) {
	char **dspstr;

        if ((dspstr=eml_setdispatch(dsp))!=(char **)0) {
	  int i;

	  for (i=0;dspstr[i]!=(char *)0;i++) {
	    (void) fprintf(fichier,"%s\n",dspstr[i]);
	    }
          dspstr=(char **)vec_freelstlst((void **)dspstr,VECFREE);
	  }
        (void) fclose(fichier);
        }
      else {
        (void) clm_alert(0,"%s unable to open file <%s> (error=<%s>)",
			    OPDI,filename,strerror(errno));
	proceed=false;	/*trouble trouble	*/
	}
      (void) free(filename);
      dsp=eml_cleandispatch(dsp);
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return completion;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to add some line to header	*/
/*						*/
/************************************************/
int eml_addheaderpart(CTTYP *cta,char **strlst)

{
#define	OPHA	"lvleml.c:eml_addheaderpart,"

int doneok;
FILE *header;
char *headername;
int phase;
int proceed;

doneok=false;
header=(FILE *)0;
(void) asprintf(&headername,"%s%s",cta->mailname,strpart[HDPART]);
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*finding header part	*/
      if ((header=fopen(headername,"a"))==(FILE *)0) {
  	(void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",OPHA,
		            headername,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*adding str line to hdr*/
      if (strlst!=(char **)0) {
	int i;

	for (i=0;strlst[i]!=(char *)0;i++) 
	  (void) fprintf(header,"%s\n",strlst[i]);
        }
      doneok=true;
      break;
    case 2	:	/*closing header file	*/
      if (fclose(header)<0) {
  	(void) clm_alert(0,"%s Unable to close <%s> (error=<%s>)",OPHA,
		            headername,strerror(errno));
	doneok=false;
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(headername);
return doneok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove the email header and*/
/*	data part.				*/
/*						*/
/************************************************/
int eml_rmemailparts(CTTYP *cta)

{
int doneok;
int i;

doneok=true;
for (i=0;i<2;i++) {
  char *filename;

  (void) asprintf(&filename,"%s%s",cta->mailname,strpart[i]);
  if (unlink(filename)<0) {
    (void) clm_alert(0,"%s Unable to remove <%s> (error=<%s>)",
		       "lvleml.c:eml_rmemailparts,",
		       filename,strerror(errno));
    doneok=false;
    }
  (void) free(filename);
  }
return doneok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to merge the email data part	*/
/*	to the email header part.		*/
/*						*/
/************************************************/
int eml_mergeemaildata(CTTYP *cta)

{
#define	OPEE	"lvleml.c:eml_mergeemaildata,"

int doneok;
FILE *store;
FILE *data;
char *dataname;
int phase;
int proceed;

doneok=false;
store=(FILE *)0;
data=(FILE *)0;
dataname=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening header file	*/	
      if ((store=fopen(cta->mailname,"w"))==(FILE *)0) {
  	(void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",OPEE,
		            cta->mailname,strerror(errno));
	proceed=false;	/*big big trouble!	*/
	}
      break;
    case 1	:	/*finding header part	*/
      (void) asprintf(&dataname,"%s%s",cta->mailname,strpart[HDPART]);
      if ((data=fopen(dataname,"r"))==(FILE *)0) {
  	(void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",OPEE,
		            dataname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*adding header part	*/
      if ((doneok=copyfile(store,data,dataname))==false)
	phase=999;	/*trouble trouble	*/
      if (fclose(data)<0) {
        doneok=false;
  	(void) clm_alert(0,"%s Unable to close <%s> (error=<%s>)",OPEE,
		              dataname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*finding data part	*/
      (void) free(dataname);
      (void) asprintf(&dataname,"%s%s",cta->mailname,strpart[DTPART]);
      if ((data=fopen(dataname,"r"))==(FILE *)0) {
  	(void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",OPEE,
		            dataname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*adding data part	*/
      if ((doneok=copyfile(store,data,dataname))==false)
	phase=999;	/*trouble trouble	*/
      if (fclose(data)<0) {
        doneok=false;
  	(void) clm_alert(0,"%s Unable to close <%s> (error=<%s>)",OPEE,
		              dataname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    default	:	/*SAFE Guard		*/
      if (fclose(store)<0) {
  	(void) clm_alert(0,"%s Unable to close <%s> (error=<%s>)",OPEE,
		            cta->mailname,strerror(errno));
	doneok=false;
	}
      (void) free(dataname);
      if (doneok==false)	/*cleaning act	*/
	(void) unlink(cta->mailname);
      proceed=false;
      break;
    }
  phase++;
  }
return doneok;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to remove the file where is	*/
/*	located trust data stream.		*/
/*						*/
/************************************************/
void eml_droptrust(CTTYP *cta)

{
char *trusname;

(void) asprintf(&trusname,"%s%s",cta->mailname,strpart[TRPART]);
if (unlink(trusname)<0) {
  (void) clm_alert(0,"%s Unable to unlink <%s> (error=<%s>)",
		     "lvleml.c:eml_droptrust",trusname,strerror(errno));
  }
(void) free(trusname);
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to extract trust data stream	*/
/*	and convert it to string array.		*/
/*						*/
/************************************************/
char **eml_gettrust(char *sessionid)

{
#define	OPGE	"lvleml.c:eml_gettrust,"

char *trusname;
char **trustlst;
FILE *store;
int phase;
int proceed;

(void) asprintf(&trusname,"%s/%s%s",clm_appdir(d_queue),sessionid,strpart[TRPART]);
trustlst=(char **)0;
store=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening the file	*/
      if ((store=fopen(trusname,"r"))==(FILE *)0) {
        (void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",
			    OPGE,trusname,strerror(errno));
	proceed=false;	/*big big big trouble	*/
	}
      break;
    case 1	:	/*reading the file	*/
      if (store!=(FILE *)0) {	/*always	*/
	char line[1024];

	while (fgets(line,sizeof(line),store)!=(char *)0) {
	  (void) clm_skiprtlf(line);
          trustlst=(char **)vec_addlstlst((void **)trustlst,strdup(line));
	  }
	(void) fclose(store);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(trusname);
return trustlst;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to get trusted data submitted	*/
/*	by remote trusted server		*/
/*						*/
/************************************************/
int eml_collecttrust(CTTYP *cta)

{
#define	OPTR	"lvleml.c:eml_collecttrust,,"

#define	TRUSLIM	(65536*16)
#define	TRUSBIG "Abort! Trust data stream too big (current %lu / maxsize=%d)"

int trustready;
u_long startmilli;
u_long stopmilli;
char *trusname;
FILE *store;
int phase;
int proceed;

trustready=false;
startmilli=(u_long)0;
stopmilli=(u_long)0;
(void) asprintf(&trusname,"%s%s",cta->mailname,strpart[TRPART]);
store=(FILE *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*opening file		*/
      if ((store=fopen(trusname,"w"))==(FILE *)0) {
	char *trouble;

        trouble=eml_ascstat(ERRPROC,"Unable to open file");
        (void) eml_printfcrlf(cta,"%s <%s>",trouble,strerror(errno));
        (void) jrl_journalling(false,EVJRL,"%s <%s> (errno=<%s>, bug?)",
                                           trouble,trusname,strerror(errno));
        (void) free(trouble);
        (void) clm_alert(0,"%s Unable to open <%s> (error=<%s>)",
			    OPTR,trusname,strerror(errno));
	proceed=false;	/*big big trouble	*/
	}
      break;
    case 1	:	/*accepting trusted data*/
      (void) eml_report(cta,TRUSTOK,"");
      (void) jrl_journalling(false,EVTRU,"%-*s%s [%s]/<%s>",6,SESSION,
					  cta->session,cta->peerip,cta->peername);
      startmilli=clm_getmillisec();
      break;
    case 2	:	/*ready to read data	*/
      phase=1;
      while (phase==1) {	/*always	*/
	char *trust;

	if ((trust=tls_waitforaction(cta->tls,cta->maxwait))==(char *)0) {
	  phase=999;	/*trouble trouble	*/
	  break;
	  }
	trust=clm_skiprtlf(trust);
	(void) fprintf(store,"%s\n",trust);
	if (strcmp(trust,DATAEND)==0) 
	  phase++;	/*stream end marker	*/
	else
          (void) jrl_journalling(false,EVTRU,"%-*s",6,trust);
	cta->mtaille+=strlen(trust);
	if (cta->mtaille>TRUSLIM) {
	  char *info;

	  (void) asprintf(&info,TRUSBIG,cta->mtaille,TRUSLIM);
	  (void) eml_report(cta,MALABRT,info);
	  (void) jrl_journalling(false,EVTRU,"%d %s",MALABRT,info);
	  (void) jrl_journalling(false,EVJRL,"%d %s",MALABRT,info);
	  (void) free(info);
	  phase=999;	/*trouble trouble	*/
	  }
	(void) free(trust);
	}
      break;
    case 3	:	/*closing file		*/
      stopmilli=clm_getmillisec();
      if (fclose(store)<0) {
        char *trouble;

        trouble=eml_ascstat(ERRPROC,"Unable to close file");
        (void) eml_printfcrlf(cta,"%s <%s>",trouble,strerror(errno));
        (void) jrl_journalling(false,EVJRL,"%s <%s> (errno=<%s>, bug?)",
                                     trouble,trusname,strerror(errno));
        (void) free(trouble);
        (void) clm_alert(0,"%s Unable to close <%s> (error=<%s>)",OPTR,
                            trusname,strerror(errno));
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*reporting stats	*/
      cta->info=clm_freestr(cta->info);
      (void) asprintf(&(cta->info),"Trusted data stream received: %ld bytes within %5.3f seconds",
				    cta->mtaille,(float)(stopmilli-startmilli)/1000.0);
      (void) jrl_journalling(false,EVJRL,"%s",cta->info);
      (void) jrl_journalling(false,EVTRU,"%s",cta->info);
      cta->info=clm_freestr(cta->info);
      cta->mtaille=0;
      trustready=true;
      break;
    default	:	/*SAFE Guard		*/
      if (trustready==false)
	(void) eml_droptrust(cta);
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(trusname);
return trustready;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to collect and store E-mail	*/
/*	data, received from remote site.	*/
/*						*/
/************************************************/
SCODTYP eml_collectdata(CTTYP *cta)

{
SCODTYP scode;
int phase;
int proceed;

scode=s_timeout;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*collecting header	*/
      if (collectheader(cta)==false) 
	phase=999;
      break;
    case 1	:	/*collecting mail data	*/
      if (collectmain(cta)==false)
	phase=999;
      break;
    case 2	:	/*preparing email file	*/
      scode=s_data;
      if (eml_mergeemaildata(cta)==false) {
	(void) eml_rmemailparts(cta);
	scode=s_unknown;
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return scode;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to grabe entry line on a TLS	*/
/*	channel, then detect DATAEND to make	*/
/*	The stream ready. if is not data-end	*/
/*	collected line are added to the stream.	*/
/************************************************/
STRMTYP *eml_collectstream(STRMTYP *stream,CTTYP *cta)

{
#define	OPCL	"lvleml.c:eml_collectstream,"

char *msg;
int phase;
int proceed;

msg=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*do we have a message?	*/
      if ((msg=tls_waitforaction(cta->tls,0))==(char *)0) 
	proceed=false;	/*no data to collect	*/
      break;
    case 1	:	/*do we have a stream	*/
      (void) clm_cleanstring(msg);
      if (stream==(STRMTYP *)0) {
	int taille;

	taille=strlen(msg);
	stream=(STRMTYP *)calloc(1,sizeof(STRMTYP));
	stream->title=calloc(taille,sizeof(char));
	stream->seq=calloc(taille,sizeof(char));
        if (sscanf(msg,"%s %s",stream->title,stream->seq)!=2) {
	  (void) clm_alert(0,"%s Unable to scan stream data <%s> (bug?)",OPCL,msg);
	  stream=eml_freestream(stream);
	  }
        (void) free(msg);
	phase=999;
	}
      break;
    case 2	:	/*is the message dataend*/
      if (strcmp(msg,DATAEND)!=0) {
        stream->msg=(char **)vec_addlstlst((void **)stream->msg,msg);
	phase=999;
	}
      break;
    case 3	:	/*stream complete	*/
      (void) free(msg);
      stream->ready=true;
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return stream;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a stream structure	*/
/*	be send.				*/
/************************************************/
STRMTYP *eml_mkstream(char *title,char *seq,char *msg)

{
STRMTYP *stream;

stream=(STRMTYP *)calloc(1,sizeof(STRMTYP));
stream->title=strdup(title);
stream->seq=strdup(seq);
if ((stream->msg=clm_splitstring(msg,'\n'))!=(char **)0) {
  while (stream->msg[stream->numline]!=(char *)0)
    stream->numline++;
  }
stream->ready=true;
return stream;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to free a stream structure	*/
/*						*/
/************************************************/
STRMTYP *eml_freestream(STRMTYP *stream)

{
if (stream!=(STRMTYP *)0) {
  stream->msg=(char **)vec_freelstlst((void **)stream->msg,VECFREE);
  stream->seq=clm_freestr(stream->seq);
  stream->title=clm_freestr(stream->title);
  (void) free(stream);
  stream=(STRMTYP *)0;
  }
return stream;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to send a data stream		*/
/*	information over an SMTP channel.	*/
/*	stream information do not require a	*/
/*	remote acknowledge...			*/
/*	Answer to request can be receive way	*/
/*	later.					*/
/*						*/
/************************************************/
int eml_sendstream(CTTYP *cta,STRMTYP *stream)

{
int status;
int proceed;
int phase;


status=true;
phase=0;
proceed=true;
if ((stream==(STRMTYP *)0)||(stream->msg==(char **)0)) {
  status=false;
  proceed=false;
  }
while (proceed==true) {
  switch (phase) {
    case   0	:
      (void) eml_printfcrlf(cta,"%s %s",stream->title,stream->seq);
      break;
    case   1	:	/*sending the stream itself	*/
      if (stream->msg!=(char **)0) {
	register int i;
	register char **ptr;

	for (i=0,ptr=stream->msg;*ptr!=(char *)0;i++,ptr++) {
	  if (strcmp(*ptr,DATAEND)==0) 
	    (void) tls_write(cta->tls,DATAEND,strlen(DATAEND));
	  (void) eml_printfcrlf(cta,"%s",*ptr);
	  }
	}
      break;
    case   2	:	/*sending the end of stream	*/
      (void) eml_printfcrlf(cta,DATAEND);
      break;
    default	:	/*safe guard			*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to post a stream on the	*/
/*	right message channel (it is a reply to	*/
/*	a previous post).			*/
/*						*/
/************************************************/
int eml_poststream(BPTR *bptr,STRMTYP *stream)

{
#define	OPPO	"lvleml.c:eml_poststream,"

int done;
char *msg;
int phase;
int proceed;

done=false;
msg=(char *)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*everything fine	*/
      if ((stream==(STRMTYP *)0)||(stream->msg==(char **)0)) {
	(void) clm_alert(0,"%s Stream or msg empty (bug?)",OPPO);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*preparing msg		*/
      if ((msg=clm_mergestring(stream->msg,'\n'))==(char *)0) {
	(void) clm_alert(0,"%s Stream msg empty! (bug?)",OPPO);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*sending message	*/
      if (msg!=(char *)0) {	/*always	*/
	TIMESPEC timeout;

	timeout.tv_sec=(time_t)2;
        timeout.tv_nsec=(long)0;
        done=true;	/*lets say it will be OK*/
        if (ipc_postreply(bptr,&timeout,stream->seq,msg)==(u_long)0) {
	  (void) clm_alert(0,"%s timeout to post reply to msgref/msg <%s/%s>",
			     OPPO,stream->seq,msg);
	  done=false;
	  phase=999;	/*trouble trouble	*/
	  }
	(void) free(msg);
	}
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return done;
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to prepare a dummy FRMTYP,	*/
/*	used when we are not able to reach	*/
/*	mail from level.			*/
/*	Case where the authentication Failed.	*/
/*						*/
/************************************************/
void eml_dummyfrm(CTTYP *contact)

{
contact->frm=(FRMTYP *)calloc(1,sizeof(FRMTYP));
contact->frm->mfrom=strdup(contact->authname);
contact->frm->grpin=contact->grpin;
contact->frm->efrom=strdup(contact->authname);
contact->frm->peerip=strdup(contact->peerip);
contact->frm->locip=strdup(contact->locip);
contact->frm->servport=strdup(contact->servport);
}
